from datetime import datetime

from shiny import App, Inputs, Outputs, Session, reactive, ui

app_ui = ui.page_fluid(
    ui.input_action_button("close", "Close the session"),
)


def server(input: Inputs, output: Outputs, session: Session):
    def log():
        print("Session ended at: " + datetime.now().strftime("%H:%M:%S"))

    session.on_ended(log)

    @reactive.effect
    @reactive.event(input.close)
    async def _():
        await session.close()


app = App(app_ui, server)



from __future__ import annotations

# Import these with underscore names so they won't show in autocomplete from the Python
# console.
from ..session import (
    Inputs as _Inputs,
    Outputs as _Outputs,
    Session as _Session,
    get_current_session as _get_current_session,
)
from .. import render
from . import ui
from ._is_express import is_express_app
from ._module import module
from ._output import (  # noqa: F401
    output_args,  # pyright: ignore[reportUnusedImport]
    suspend_display,  # pyright: ignore[reportUnusedImport] - Deprecated
)
from ._run import app_opts, wrap_express_app
from .expressify_decorator import expressify


__all__ = (
    "render",
    "input",
    "output",
    "session",
    "is_express_app",
    "app_opts",
    "wrap_express_app",
    "ui",
    "expressify",
    "module",
)

# Add types to help type checkers
input: _Inputs
output: _Outputs
session: _Session


# Note that users should use `from shiny.express import input` instead of `from shiny
# import express` and acces via `express.input`. The former provides a static value for
# `input`, but the latter is dynamic -- every time `express.input` is accessed, it
# returns the input for the current session. This will work in the vast majority of
# cases, but when it fails, it will be very confusing.
def __getattr__(name: str) -> object:
    session = _get_current_session()

    if name == "input":
        if session is None:
            return _ExpressOnlyPlaceholder("input")
        else:
            return session.input  # pyright: ignore
    elif name == "output":
        if session is None:
            return _ExpressOnlyPlaceholder("output")
        else:
            return session.output  # pyright: ignore
    elif name == "session":
        if session is None:
            return _ExpressOnlyPlaceholder("session")
        else:
            return session

    raise AttributeError(f"Module 'shiny.express' has no attribute '{name}'")


class _ExpressOnlyPlaceholder:
    """Placeholder class for objects that can only be used in a Shiny Express app."""

    def __init__(self, name: str):
        self.name = name

    def __getattr__(self, name: str) -> None:
        raise RuntimeError(
            f"shiny.express.{self.name} can only be used inside of a Shiny Express app."
        )

    def __call__(self, *args: object, **kwargs: object) -> None:
        raise RuntimeError(
            f"shiny.express.{self.name} can only be used inside of a Shiny Express app."
        )



from palmerpenguins import load_penguins_raw  # pyright: ignore[reportMissingTypeStubs]

from shiny import App, Inputs, Outputs, Session, render, ui
from shiny.render import CellPatch
from shiny.types import SafeException

app_ui = ui.page_fluid(
    ui.h2("Palmer Penguins"),
    ui.output_data_frame("penguins_df"),
)


def server(input: Inputs, output: Outputs, session: Session) -> None:
    @render.data_frame
    def penguins_df():
        return render.DataGrid(
            data=load_penguins_raw().head(
                5
            ),  # pyright: ignore[reportUnknownArgumentType]
            filters=True,
            editable=True,
        )

    @penguins_df.set_patch_fn
    async def upgrade_patch(
        *,
        patch: CellPatch,
    ):
        if (patch["column_index"] == 4) and (
            patch["value"] not in ["Torgersen", "Biscoe", "Dream"]
        ):  # check island
            raise SafeException(
                "Penguin island should be one of 'Torgersen', 'Biscoe', 'Dream'"
            )
        if (patch["column_index"] == 9) and int(
            patch["value"]  # pyright: ignore[reportArgumentType]
        ) > 50:  # check culmen length
            raise SafeException("Penguin culmen length cannot be greater than 50mm")
        return patch["value"]


app = App(app_ui, server)



import numpy as np
from plotly import graph_objects as go

# Code taken from https://plotly.com/python/3d-line-plots/
# * `brownian_motion()` is a function that generates a random walk
# * `brownian_widget()` uses restructured plotting code given in article

rs = np.random.RandomState()


# https://plotly.com/python/3d-line-plots/
def brownian_motion(T=1, N=100, mu=0.1, sigma=0.01, S0=20):
    dt = float(T) / N
    t = np.linspace(0, T, N)
    W = rs.standard_normal(size=N)
    W = np.cumsum(W) * np.sqrt(dt)  # standard brownian motion
    X = (mu - 0.5 * sigma**2) * t + sigma * W
    S = S0 * np.exp(X)  # geometric brownian motion
    return S


def brownian_data(n=100, mu=(0.0, 0.00), sigma=(0.1, 0.1), S0=(1.0, 1.0)):
    return {
        "x": brownian_motion(T=1, N=n, mu=mu[0], sigma=sigma[0], S0=S0[0]),
        "y": brownian_motion(T=1, N=n, mu=mu[1], sigma=sigma[1], S0=S0[1]),
        # "y": [i for i in range(n)],
        "z": [i for i in range(n)],
    }


def brownian_widget(width=600, height=600):
    widget = go.FigureWidget(
        data=[
            go.Scatter3d(
                x=[],
                y=[],
                z=[],
                marker=dict(
                    size=4,
                    color=[],
                    colorscale="Viridis",
                ),
                line=dict(color="darkblue", width=2),
            )
        ],
        layout={"showlegend": False, "width": width, "height": height},
    )

    return widget



import re

import pytest
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.render._data_frame_utils._types import (
    ColumnFilterNumber,
    ColumnFilterStr,
    ColumnSort,
)
from shiny.run import ShinyAppProc


@pytest.mark.parametrize("tab_name", ["pandas", "polars"])
def test_validate_html_columns(
    page: Page,
    local_app: ShinyAppProc,
    tab_name: str,
) -> None:
    page.goto(local_app.url)

    data_frame = controller.OutputDataFrame(page, f"{tab_name}-penguins_df")
    tab = controller.NavsetUnderline(page, "tab")
    tab.set(tab_name)

    controller.OutputTextVerbatim(page, f"{tab_name}-df_type").expect_value(
        re.compile(tab_name)
    )

    sort = controller.OutputTextVerbatim(page, f"{tab_name}-sort")
    filter = controller.OutputTextVerbatim(page, f"{tab_name}-filter")
    rows = controller.OutputTextVerbatim(page, f"{tab_name}-rows")
    selected_rows = controller.OutputTextVerbatim(page, f"{tab_name}-selected_rows")

    sort.expect_value("()")
    filter.expect_value("()")
    rows.expect_value("(0, 1, 2, 3, 4)")
    selected_rows.expect_value("()")
    col: ColumnSort = {
        "col": 2,
        "desc": True,
    }
    data_frame.set_sort(col)
    sort.expect_value("({'col': 2, 'desc': True},)")
    filter.expect_value("()")
    rows.expect_value("(2, 3, 4, 0, 1)")
    selected_rows.expect_value("()")
    data_frame.select_rows([1, 3])
    selected_rows.expect_value("(3, 0)")
    data_frame.select_rows([1, 3])  # unselect the rows
    selected_rows.expect_value("()")

    filter_text: ColumnFilterStr = {"col": 1, "value": "A2"}
    data_frame.set_filter(filter_text)
    sort.expect_value("({'col': 2, 'desc': True},)")

    filter_criteria_num: ColumnFilterNumber = {"col": 0, "value": [2, None]}
    data_frame.set_filter([filter_criteria_num])
    filter.expect_value("({'col': 0, 'value': (2, None)},)")
    data_frame.set_filter([filter_criteria_num, filter_text])
    filter.expect_value("({'col': 0, 'value': (2, None)}, {'col': 1, 'value': 'A2'})")

    rows.expect_value("(3, 1)")
    selected_rows.expect_value("()")
    data_frame.select_rows([0])
    rows.expect_value("(3, 1)")
    selected_rows.expect_value("(3,)")



import pandas as pd

from shiny import App, Inputs, render, ui

green_styles = [
    {
        "rows": [2, 4],
        "cols": [2, 4],
        "style": {
            "background-color": "mediumspringgreen",
            "width": "300px",
            "height": "100px",
        },
    }
]

n = 6
df = pd.DataFrame(
    {
        "a": range(n),
        "b": range(n, n * 2),
        "c": range(n * 2, n * 3),
        "d": range(n * 3, n * 4),
        "e": range(n * 4, n * 5),
    }
)

hi_styles = [
    {
        # No `rows` or `cols` means apply to all cells
        "class": "posit-bg",
        "style": {
            "border": "transparent",
            "color": "transparent",
        },
    },
    {
        "rows": [3],
        "cols": [2],
        "class": "posit-blue-bg",
        "style": {
            "width": "100px",
            "height": "75px",
        },
    },
    {
        "cols": [1, 3, 5],
        "class": "posit-blue-bg",
    },
    {
        "cols": [7],
        "rows": [0, 1, 2, 3, 5],
        "class": "posit-orange-bg",
    },
]

n = 7
hi_pd = pd.DataFrame(
    {
        "a": range(n),
        "b": range(n, n * 2),
        "c": range(n * 2, n * 3),
        "d": range(n * 3, n * 4),
        "e": range(n * 4, n * 5),
        "f": range(n * 5, n * 6),
        "g": range(n * 6, n * 7),
        "h": range(n * 7, n * 8),
        "i": range(n * 8, n * 9),
    }
)


app_ui = ui.page_fillable(
    ui.h2("Data Frame with Styles applied to 4 cells"),
    ui.output_data_frame("my_df"),
    ui.hr(),
    ui.tags.style(
        ui.HTML(
            """
        .posit-bg {
            background-color: #242a26 ;
        }
        .posit-blue-bg {
            background-color: #447099 ;
        }
        .posit-orange-bg {
            background-color: #ED642F ;
        }
        """
        )
    ),
    ui.h2(
        "Custom styles applied to all cells within a data frame ", ui.HTML("&#128075;")
    ),
    ui.output_data_frame("hi_df"),
)


def server(input: Inputs):
    @render.data_frame
    def my_df():
        return render.DataGrid(
            df,
            styles=green_styles,
        )

    @render.data_frame
    def hi_df():
        return render.DataGrid(
            hi_pd,
            styles=hi_styles,
        )


app = App(app_ui, server)



import os

import pytest
from playwright.sync_api import Page
from utils.deploy_utils import (
    local_deploys_app_url_fixture,
    reruns,
    reruns_delay,
    skip_if_not_chrome,
)

from shiny.playwright import controller

app_url = local_deploys_app_url_fixture(
    # Possibly use a different app name given by an GHA env var
    os.getenv("EXPRESS_PAGE_SIDEBAR_NAME")
    or "express_page_sidebar"
)


@skip_if_not_chrome
@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
def test_express_page_sidebar(page: Page, app_url: str) -> None:
    page.goto(app_url)

    sidebar = controller.Sidebar(page, "sidebar")
    sidebar.expect_text("SidebarTitle Sidebar Content")
    output_txt = controller.OutputTextVerbatim(page, "txt")
    output_txt.expect_value("50")



from playwright.sync_api import Page
from playwright.sync_api import expect as playwright_expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_input_action_link_kitchen(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    default = controller.InputActionLink(page, "default")
    default.expect_label("Default action link")
    playwright_expect(default.loc.locator("svg.fa")).to_have_count(0)
    controller.OutputCode(page, "default_txt").expect_value("Link clicked 0 times")
    default.click()
    controller.OutputCode(page, "default_txt").expect_value("Link clicked 1 times")

    icon = controller.InputActionLink(page, "icon")
    playwright_expect(icon.loc.locator("svg.fa")).to_have_count(1)



from shiny import App, Inputs, Outputs, Session, reactive, ui


def run_model(delay=10.0):
    import time

    # Pretend to run a model for `delay` seconds
    start_time = time.time()
    while time.time() - start_time < delay:
        pass
    return time.time()


def the_modal():
    return ui.modal(
        "The model is running, please wait.",
        title="Running model",
        easy_close=False,
        footer=None,
    )


app_ui = ui.page_fluid(
    ui.input_action_button("run", "Run Model"),
)


def server(input: Inputs, output: Outputs, session: Session):
    model_result = reactive.value()

    @reactive.effect
    @reactive.event(input.run)
    def do_run_model():
        # Show the modal, blocking interaction with the UI
        ui.modal_show(the_modal())

        result = run_model(delay=4)

        # Now that we have model results, remove the modal
        # and update the model result reactive value
        ui.modal_remove()
        model_result.set(result)


app = App(app_ui, server)



from shiny import App, Inputs, Outputs, Session, reactive, ui

app_ui = ui.page_fluid(
    ui.input_action_button("btn_show", "Show tooltip", class_="mt-3 me-3"),
    ui.input_action_button("btn_close", "Close tooltip", class_="mt-3 me-3"),
    ui.input_action_button(
        "btn_update", "Update tooltip phrase (and show tooltip)", class_="mt-3 me-3"
    ),
    ui.tooltip(
        ui.input_action_button(
            "btn_w_tooltip", "A button w/ a tooltip", class_="btn-primary mt-5"
        ),
        "A message",
        id="tooltip_id",
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    @reactive.event(input.btn_show)
    def _():
        ui.update_tooltip("tooltip_id", show=True)

    @reactive.effect
    @reactive.event(input.btn_close)
    def _():
        ui.update_tooltip("tooltip_id", show=False)

    @reactive.effect
    @reactive.event(input.btn_update)
    def _():
        content = (
            "A " + " ".join(["NEW" for _ in range(input.btn_update())]) + " message"
        )

        ui.update_tooltip("tooltip_id", content, show=True)

    @reactive.effect
    @reactive.event(input.btn_w_tooltip)
    def _():
        ui.notification_show("Button clicked!", duration=3, type="message")


app = App(app_ui, server=server)



from shiny import reactive
from shiny.express import input, render, ui

with ui.sidebar(id="sidebar"):
    "Sidebar content"

ui.input_action_button("open_sidebar", label="Open sidebar", class_="me-3")
ui.input_action_button("close_sidebar", label="Close sidebar", class_="me-3")


@render.text
def state():
    return f"input.sidebar(): {input.sidebar()}"


@reactive.effect
@reactive.event(input.open_sidebar)
def _():
    ui.update_sidebar("sidebar", show=True)


@reactive.effect
@reactive.event(input.close_sidebar)
def _():
    ui.update_sidebar("sidebar", show=False)



from __future__ import annotations

import pytest
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.playwright import expect as shiny_expect
from shiny.run import ShinyAppProc
from shiny.types import Jsonifiable


@pytest.mark.parametrize("tab_name", ["pandas", "polars"])
def test_validate_column_labels(
    page: Page, local_app: ShinyAppProc, tab_name: str
) -> None:
    page.goto(local_app.url)

    controller.NavsetUnderline(page, "tab").set(tab_name)

    fn_styles = controller.OutputDataFrame(page, f"{tab_name}-fn_styles")
    list_styles = controller.OutputDataFrame(page, f"{tab_name}-list_styles")

    styles: list[dict[str, Jsonifiable]] = [
        {
            "rows": [1, 2],
            "cols": [2],  # "Species",
            "style": {
                "background-color": "purple",
                "border-color": "green",
                "border-style": "dashed",
            },
        },
        {
            "rows": [2],
            "cols": [3],  # "Region",
            "style": {"background-color": "yellow"},
        },
        {
            "rows": None,
            "cols": [4],  # "Island",
            "style": {"background-color": "red"},
        },
        {
            "rows": [1],
            "cols": [4, 5],  # "Island", "Stage",
            "style": {"background-color": "green"},
        },
    ]

    every_styles: dict[str, Jsonifiable] = {
        "location": "body",
        "rows": None,
        "cols": None,
        "style": {"background-color": "lightblue", "color": "darkorange"},
    }
    every_styles_with_style: list[dict[str, Jsonifiable]] = [every_styles]
    for style in styles:
        every_styles_with_style.append(style)

    def expect_styles(
        df: controller.OutputDataFrame, styles: list[dict[str, Jsonifiable]]
    ):
        def style_for_cell(row: int, col: int) -> dict[str, Jsonifiable] | None:
            last_val: dict[str, Jsonifiable] | None = None
            for style in styles:
                if isinstance(style["rows"], list):
                    if row not in style["rows"]:
                        continue
                if isinstance(style["cols"], list):
                    if col not in style["cols"]:
                        continue
                assert isinstance(style["style"], dict)
                last_val = style["style"]
            return last_val

        for row in range(0, 5):
            for col in range(0, 6):
                cell_loc = df.cell_locator(row=row, col=col)
                ex_cell_style = style_for_cell(row, col)

                if ex_cell_style is None:

                    shiny_expect.expect_not_to_have_style(cell_loc, "background-color")
                else:
                    assert isinstance(ex_cell_style, dict)
                    assert "background-color" in ex_cell_style
                    assert isinstance(ex_cell_style["background-color"], str)

                    shiny_expect.expect_to_have_style(
                        cell_loc, "background-color", ex_cell_style["background-color"]
                    )

    expect_styles(fn_styles, [styles[0]])
    expect_styles(list_styles, every_styles_with_style)

    fn_styles.set_cell("new value", row=0, col=0)
    expect_styles(fn_styles, [styles[0], styles[1]])

    fn_styles.set_cell("new value2", row=1, col=0)
    expect_styles(fn_styles, [styles[0], styles[1], styles[2]])

    fn_styles.set_cell("new value3", row=2, col=0)
    expect_styles(fn_styles, [styles[0], styles[1], styles[2], styles[3]])

    fn_styles.set_cell("new value4", row=3, col=0)
    expect_styles(fn_styles, [styles[0]])



from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("input_numeric")


# def shinypage(page: Page) -> Page:
#     # look up the pytest.mark.shinyapp, throw if it's not there
#     # Use conftest's run_shiny_app to start the app
#     with run_shiny_app(...) as app:
#         if navigate:
#             page.goto(app.url)
#         yield page
#
#
# @pytest.mark.shinyapp(example="input_numeric", navigate=False)
# def test_input_numeric_tmp(shinypage: Page) -> None:
#     ...


def test_input_numeric_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    # page.set_default_timeout(1000)

    obs = controller.InputNumeric(page, "obs")
    obs.expect.to_have_value("10")
    expect(obs.loc).to_have_value("10")

    expect(obs.loc_label).to_have_text("Observations:")

    obs.expect_label("Observations:")
    obs.expect_value("10")
    obs.expect_min("1")
    obs.expect_max("100")
    obs.expect_step(None)
    obs.expect_width(None)


def test_input_numeric_typical(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    # page.set_default_timeout(1000)

    obs = controller.InputNumeric(page, "obs")
    obs.expect.to_have_value("10")
    obs.loc.fill("42")
    obs.expect.not_to_have_value("10")
    obs.expect.to_have_value("42")

    expect(obs.loc).to_have_value("42")


def test_input_numeric_app(page: Page, app: ShinyAppProc) -> None:
    # with page and app:
    page.goto(app.url)

    obs = controller.InputNumeric(page, "obs")
    # obs.label.expect.to_have_text("Observed")
    obs.expect.to_have_value("10")

    value = controller.OutputTextVerbatim(page, "value")
    value.expect_value("10")

    output = page.locator("#value")
    expect(output).to_have_text("10")

    obs.loc.fill("21")
    obs.expect.to_have_value("21")
    value.expect_value("21")



import matplotlib.pyplot as plt
import numpy as np

from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.layout_sidebar(
        ui.sidebar(
            ui.input_slider("n", "N", min=0, max=100, value=20),
        ),
        ui.output_plot("plot"),
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.plot(alt="A histogram")
    def plot() -> object:
        np.random.seed(19680801)
        x = 100 + 15 * np.random.randn(437)

        fig, ax = plt.subplots()
        ax.hist(x, input.n(), density=True)
        return fig


app = App(app_ui, server)



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_output_image_kitchen(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    btn_count = controller.InputActionButton(page, "btn_count")
    btn_trigger = controller.InputActionButton(page, "btn_trigger")
    txt_immediate = controller.OutputTextVerbatim(page, "txt_immediate")
    txt_render_delayed = controller.OutputTextVerbatim(page, "txt_render_delayed")
    txt_reactive_delayed = controller.OutputTextVerbatim(page, "txt_reactive_delayed")

    txt_immediate.expect_value("0")
    txt_render_delayed.expect_value("")
    txt_reactive_delayed.expect_value("")

    btn_count.click()
    btn_count.click()
    btn_count.click()
    txt_immediate.expect_value("3")
    txt_render_delayed.expect_value("")
    txt_reactive_delayed.expect_value("")

    btn_trigger.click()
    txt_immediate.expect_value("3")
    txt_render_delayed.expect_value("3")
    txt_reactive_delayed.expect_value("3")

    btn_count.click()
    btn_count.click()
    txt_immediate.expect_value("5")
    txt_render_delayed.expect_value("3")
    txt_reactive_delayed.expect_value("3")

    btn_trigger.click()
    txt_immediate.expect_value("5")
    txt_render_delayed.expect_value("5")
    txt_reactive_delayed.expect_value("5")



# Generated by scripts/htmlDependencies.R: do not edit by hand

from __future__ import annotations

from typing import Literal

ShinyThemePreset = Literal[
    "bootstrap",
    "shiny",
    "cerulean",
    "cosmo",
    "cyborg",
    "darkly",
    "flatly",
    "journal",
    "litera",
    "lumen",
    "lux",
    "materia",
    "minty",
    "morph",
    "pulse",
    "quartz",
    "sandstone",
    "simplex",
    "sketchy",
    "slate",
    "solar",
    "spacelab",
    "superhero",
    "united",
    "vapor",
    "yeti",
    "zephyr",
]

shiny_theme_presets: tuple[ShinyThemePreset, ...] = (
    "bootstrap",
    "shiny",
    "cerulean",
    "cosmo",
    "cyborg",
    "darkly",
    "flatly",
    "journal",
    "litera",
    "lumen",
    "lux",
    "materia",
    "minty",
    "morph",
    "pulse",
    "quartz",
    "sandstone",
    "simplex",
    "sketchy",
    "slate",
    "solar",
    "spacelab",
    "superhero",
    "united",
    "vapor",
    "yeti",
    "zephyr",
)

shiny_theme_presets_bundled: tuple[ShinyThemePreset, ...] = (
    "bootstrap",
    "shiny",
)

shiny_theme_presets_bootswatch: tuple[ShinyThemePreset, ...] = (
    "cerulean",
    "cosmo",
    "cyborg",
    "darkly",
    "flatly",
    "journal",
    "litera",
    "lumen",
    "lux",
    "materia",
    "minty",
    "morph",
    "pulse",
    "quartz",
    "sandstone",
    "simplex",
    "sketchy",
    "slate",
    "solar",
    "spacelab",
    "superhero",
    "united",
    "vapor",
    "yeti",
    "zephyr",
)



from datetime import date, timedelta

from shiny import reactive
from shiny.express import input, ui

ui.input_slider("n", "Day of month", min=1, max=30, value=10)
ui.input_date("inDate", "Input date")


@reactive.effect
def _():
    d = date(2013, 4, input.n())
    ui.update_date(
        "inDate",
        label="Date label " + str(input.n()),
        value=d,
        min=d - timedelta(days=3),
        max=d + timedelta(days=3),
    )



from __future__ import annotations

__all__ = (
    "modal_button",
    "modal",
    "modal_show",
    "modal_remove",
)

from typing import TYPE_CHECKING, Literal, Optional

from htmltools import HTML, Tag, TagAttrs, TagAttrValue, TagChild, div, tags

from .._docstring import add_example
from ..session import require_active_session
from ..types import MISSING, MISSING_TYPE

if TYPE_CHECKING:
    from ..session import Session


@add_example(ex_dir="../api-examples/modal")
def modal_button(label: TagChild, icon: TagChild = None, **kwargs: TagAttrValue) -> Tag:
    """
    Creates a button that will dismiss a :func:`~shiny.ui.modal`.

    :func:`~shiny.ui.modal_button` is usually
    passed to the `footer` of a :func:`~shiny.ui.modal` to add a button to the footer that will close
    the :func:`~shiny.ui.modal`.

    Parameters
    ----------
    label
        An input label.
    icon
        An icon to appear inline with the button/link.
    **kwargs
        Attributes to be applied to the button.

    Returns
    -------
    :
        A UI element

    See Also
    --------
    * :func:`~shiny.ui.modal`
    * :func:`~shiny.ui.modal_show`
    * :func:`~shiny.ui.modal_remove`
    """
    return tags.button(
        icon,
        label,
        {"class": "btn btn-default"},
        type="button",
        data_dismiss="modal",
        data_bs_dismiss="modal",
        **kwargs,
    )


@add_example()
def modal(
    *args: TagChild | TagAttrs,
    title: Optional[str] = None,
    footer: TagChild | MISSING_TYPE = MISSING,
    size: Literal["m", "s", "l", "xl"] = "m",
    easy_close: bool = False,
    fade: bool = True,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Creates the UI for a modal dialog, using Bootstrap's modal class.

    A modal is a dialog box that appears in front of the app. Modals are
    typically used for showing important messages, or for presenting UI that requires
    input from the user, such as a user name and/or password input.

    Parameters
    ----------
    *args
        UI elements for the body of the modal.
    title
        An optional title for the modal dialog.
    footer
        UI for footer. Use ``None`` for no footer.
    size
        The size of the modal dialogue box. Use one of "s" for small, "m" (the default)
        for medium, or "l" for large.
    easy_close
        If ``True``, the modal dialog can be dismissed by clicking outside the dialog
        box, or by pressing the Escape key. If ``False`` (the default), the modal dialog
        can't be dismissed in those ways; instead it must be dismissed by clicking on a
        ``modal_button()``, or from a call to ``modal_remove()`` on the server.
    fade
        If ``False``, the modal dialog will have no fade-in animation (it will simply
        appear rather than fade in to view).
    **kwargs
        Attributes to be applied to the modal's body tag.

    Returns
    -------
    :
        A UI element

    See Also
    --------
    * :func:`~shiny.ui.modal_show`
    * :func:`~shiny.ui.modal_remove`
    * :func:`~shiny.ui.modal_button`
    """

    title_div = None
    if title:
        title_div = div(tags.h4(title, class_="modal-title"), class_="modal-header")

    if isinstance(footer, MISSING_TYPE):
        footer = modal_button("Dismiss")
    if footer is not None:
        footer = div(footer, class_="modal-footer")

    dialog = div(
        div(
            title_div,
            div({"class": "modal-body"}, *args, **kwargs),
            footer,
            class_="modal-content",
        ),
        class_="modal-dialog"
        + ({"s": " modal-sm", "l": " modal-lg", "xl": " modal-xl"}.get(size, "")),
    )

    # jQuery plugin doesn't work in Bootstrap 5, but vanilla JS doesn't work in Bootstrap 4 :sob:
    js = "\n".join(
        [
            "if (window.bootstrap && !window.bootstrap.Modal.VERSION.match(/^4\\. /)) {",
            "  var modal=new bootstrap.Modal(document.getElementById('shiny-modal'))",
            "  modal.show()",
            "} else {",
            "  $('#shiny-modal').modal().focus()",
            "}",
        ]
    )

    backdrop = None if easy_close else "static"
    keyboard = None if easy_close else "false"

    return div(
        dialog,
        tags.script(HTML(js)),
        id="shiny-modal",
        class_="modal fade" if fade else "modal",
        tabindex="-1",
        data_backdrop=backdrop,
        data_bs_backdrop=backdrop,
        data_keyboard=keyboard,
        data_bs_keyboard=keyboard,
    )


@add_example(ex_dir="../api-examples/modal")
def modal_show(modal: Tag, session: Optional[Session] = None) -> None:
    """
    Show a modal dialog.

    :func:`~shiny.ui.modal_show` is used to display a modal that has been
    created with :func:`~shiny.ui.modal`.

    Parameters
    ----------
    modal
        Typically a :func:`~shiny.ui.modal` instance.
    session
        The :class:`~shiny.Session` instance to display the modal in. If not provided,
        the session is inferred via :func:`~shiny.session.get_current_session`.

    See Also
    --------
    * :func:`~shiny.ui.modal_remove`
    * :func:`~shiny.ui.modal`
    """
    session = require_active_session(session)
    msg = session._process_ui(modal)
    session._send_message_sync({"modal": {"type": "show", "message": msg}})


@add_example()
def modal_remove(session: Optional[Session] = None) -> None:
    """
    Remove a modal dialog box.

    :func:`~shiny.ui.modal_remove` provides a way to remove a modal programmatically.
    Modals can also be removed manually by the user if a :func:`~shiny.ui.modal_button`
    is provided, or if the modal is created with `easy_close=True`.

    Parameters
    ----------
    session
        The :class:`~shiny.Session` instance that contains the modal to remove. If not
        provided, the session is inferred via :func:`~shiny.session.get_current_session`.

    See Also
    --------
    * :func:`~shiny.ui.modal_show`
    * :func:`~shiny.ui.modal`
    """
    session = require_active_session(session)
    session._send_message_sync({"modal": {"type": "remove", "message": None}})



from shiny import reactive
from shiny.express import input, ui

ui.input_action_button("set_acc", "Only open sections A,C,E", class_="mt-3 mb-3")

with ui.accordion(id="acc", open=["Section B", "Section D"], multiple=True):
    for letter in "ABCDE":
        with ui.accordion_panel(f"Section {letter}"):
            f"Some narrative for section {letter}"


@reactive.effect
@reactive.event(input.set_acc)
def _():
    ui.update_accordion("acc", show=["Section A", "Section C", "Section E"])



__all__ = ("download_button", "download_link")

from typing import Optional

from htmltools import Tag, TagAttrValue, TagChild, css, tags

from .._docstring import add_example
from .._namespaces import resolve_id
from .._shinyenv import is_pyodide


@add_example()
def download_button(
    id: str,
    label: TagChild,
    *,
    icon: TagChild = None,
    width: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Create a download button

    Parameters
    ----------
    id
        An id for the download.
    label
        An input label.
    icon
        An icon to display on the button.
    width
        The width of the button.
    **kwargs
        Additional attributes for the button.

    Returns
    -------
    :
        A UI element

    See Also
    --------
    * :class:`~shiny.render.download`
    * :func:`~shiny.ui.download_link`
    """

    return tags.a(
        icon,
        label,
        {
            "class": "btn btn-default shiny-download-link disabled",
            "style": css(width=width),
        },
        id=resolve_id(id),
        href="",
        target="_blank",
        # We can't use `download` in pyodide mode, because the browser chooses not to
        # route the download through the service worker in that case. (Observed by
        # jcheng on 1/7/2022, using Chrome Version 96.0.4664.110.)
        download=None if is_pyodide else True,
        aria_disabled="true",
        tabindex="-1",
        **kwargs,
    )


@add_example()
def download_link(
    id: str,
    label: TagChild,
    *,
    icon: TagChild = None,
    width: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Create a download button.

    Parameters
    ----------
    id
        An id for the download.
    label
        An input label.
    icon
        An icon to display on the button.
    width
        The width of the button.
    **kwargs
        Additional attributes for the button.

    Returns
    -------
    :
        A UI element

    See Also
    --------
    * :class:`~shiny.render.download`
    * :func:`~shiny.ui.download_button`
    """

    return tags.a(
        icon,
        label,
        {"class": "shiny-download-link disabled", "style": css(width=width)},
        id=resolve_id(id),
        href="",
        target="_blank",
        download=None if is_pyodide else True,
        aria_disabled="true",
        tabindex="-1",
        **kwargs,
    )



"Context manager components for Shiny Express"

from __future__ import annotations

from typing import Literal, Optional

from htmltools import Tag, TagAttrs, TagAttrValue, TagChild, TagFunction, TagList

from ... import ui
from ..._docstring import add_example, no_example
from ...types import MISSING, MISSING_TYPE
from ...ui._accordion import AccordionPanel
from ...ui._card import CardItem
from ...ui._layout_columns import BreakpointsUser
from ...ui._navs import NavMenu, NavPanel, NavSet, NavSetBar, NavSetCard
from ...ui._sidebar import SidebarOpenSpec, SidebarOpenValue
from ...ui.css import CssUnit
from .._recall_context import RecallContextManager

__all__ = (
    "sidebar",
    "layout_sidebar",
    "layout_column_wrap",
    "layout_columns",
    "card",
    "card_body",
    "card_header",
    "card_footer",
    "accordion",
    "accordion_panel",
    "nav_panel",
    "nav_control",
    "nav_menu",
    "panel_well",
    "panel_conditional",
    "panel_fixed",
    "panel_absolute",
)


# ======================================================================================
# Shiny layout components
# ======================================================================================
@add_example()
def sidebar(
    *,
    position: Literal["left", "right"] = "left",
    open: Optional[SidebarOpenSpec | SidebarOpenValue | Literal["desktop"]] = None,
    width: CssUnit = 250,
    id: Optional[str] = None,
    title: TagChild | str = None,
    bg: Optional[str] = None,
    fg: Optional[str] = None,
    class_: Optional[str] = None,
    max_height_mobile: Optional[str | float] = None,
    gap: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    **kwargs: TagAttrValue,
) -> RecallContextManager[ui.Sidebar]:
    """
    Context manager for sidebar element

    This function wraps :func:`~shiny.ui.sidebar`.

    Parameters
    ----------
    width
        A valid CSS unit used for the width of the sidebar.
    position
        Where the sidebar should appear relative to the main content, one of `"left"` or
        `"right"`.
    open
        The initial state of the sidebar. If a string, the possible values are:

        * `"open"`: the sidebar starts open
        * `"closed"`: the sidebar starts closed
        * `"always"`: the sidebar is always open and cannot be closed

        Alternatively, you can provide a dictionary with keys `"desktop"` and `"mobile"`
        to set different initial states for desktop and mobile. For example, when
        `{"desktop": "open", "mobile": "closed"}` the sidebar is initialized in the
        open state on desktop screens or in the closed state on mobile screens.
    id
        A character string. Required if wanting to reactively read (or update) the
        `collapsible` state in a Shiny app.
    title
        A character title to be used as the sidebar title, which will be wrapped in a
        `<div>` element with class `sidebar-title`. You can also provide a custom
        :class:`~htmltools.Tag` for the title element, in which case you'll
        likely want to give this element `class = "sidebar-title"`.
    bg,fg
        A background or foreground color.
    class_
        CSS classes for the sidebar container element, in addition to the fixed
        `.sidebar` class.
    max_height_mobile
        A CSS length unit (passed through :func:`~shiny.ui.css.as_css_unit`) defining
        the maximum height of the horizontal sidebar when viewed on mobile devices. Only
        applies to always-open sidebars on mobile, where by default the sidebar
        container is placed below the main content container on mobile devices.
    gap
        A CSS length unit defining the vertical `gap` (i.e., spacing) between elements
        provided to `*args`.
    padding
        Padding within the sidebar itself. This can be a numeric vector (which will be
        interpreted as pixels) or a character vector with valid CSS lengths. `padding`
        may be one to four values.

        * If a single value, then that value will be used for all four sides.
        * If two, then the first value will be used for the top and bottom, while
          the second value will be used for left and right.
        * If three values, then the first will be used for top, the second will be left
          and right, and the third will be bottom.
        * If four, then the values will be interpreted as top, right, bottom, and left
          respectively.
    **kwargs
        Named attributes are supplied to the sidebar content container.
    """
    return RecallContextManager(
        ui.sidebar,
        kwargs=dict(
            width=width,
            position=position,
            open=open,
            id=id,
            title=title,
            bg=bg,
            fg=fg,
            class_=class_,
            max_height_mobile=max_height_mobile,
            gap=gap,
            padding=padding,
            **kwargs,
        ),
    )


# TODO: Figure out sidebar arg for ui.layout_sidebar
@add_example()
def layout_sidebar(
    *,
    fillable: bool = True,
    fill: bool = True,
    bg: Optional[str] = None,
    fg: Optional[str] = None,
    border: Optional[bool] = None,
    border_radius: Optional[bool] = None,
    border_color: Optional[str] = None,
    gap: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    height: Optional[CssUnit] = None,
    **kwargs: TagAttrValue,
) -> RecallContextManager[CardItem]:
    """
    Context manager for sidebar layout

    This function wraps :func:`~shiny.ui.layout_sidebar`.

    Create a sidebar layout component which can be dropped inside any Shiny UI page
    method (e.g. :func:`~shiny.ui.page_fillable`) or :func:`~shiny.ui.card`
    context.

    The first child needs to be of class :class:`~shiny.ui.Sidebar` object created by
    :func:`~shiny.express.ui.sidebar`. The remaining arguments will contain the contents
    to the main content area. Or tag attributes that are supplied to the resolved
    :class:`~htmltools.Tag` object.

    Parameters
    ----------
    fillable
        Whether or not the main content area should be wrapped in a fillable container.
        See :func:`~shiny.ui.fill.as_fillable_container` for details.
    fill
        Whether or not the sidebar layout should be wrapped in a fillable container. See
        :func:`~shiny.ui.fill.as_fill_item` for details.
    bg,fg
        A background or foreground color.
    border
        Whether or not to show a border around the sidebar layout.
    border_radius
        Whether or not to round the corners of the sidebar layout.
    border_color
        A border color.
    gap
        A CSS length unit defining the vertical `gap` (i.e., spacing) between elements
        provided to `*args`. This value will only be used if `fillable` is `True`.
    padding
        Padding within the sidebar itself. This can be a numeric vector (which will be
        interpreted as pixels) or a character vector with valid CSS lengths. `padding`
        may be one to four values. If one, then that value will be used for all four
        sides. If two, then the first value will be used for the top and bottom, while
        the second value will be used for left and right. If three, then the first will
        be used for top, the second will be left and right, and the third will be
        bottom. If four, then the values will be interpreted as top, right, bottom, and
        left respectively.
    height
        Any valid CSS unit to use for the height.
    """
    return RecallContextManager(
        ui.layout_sidebar,
        kwargs=dict(
            fillable=fillable,
            fill=fill,
            bg=bg,
            fg=fg,
            border=border,
            border_radius=border_radius,
            border_color=border_color,
            gap=gap,
            padding=padding,
            height=height,
            **kwargs,
        ),
    )


@add_example()
def layout_column_wrap(
    *,
    width: CssUnit | None | MISSING_TYPE = MISSING,
    fixed_width: bool = False,
    heights_equal: Literal["all", "row"] = "all",
    fill: bool = True,
    fillable: bool = True,
    height: Optional[CssUnit] = None,
    min_height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    height_mobile: Optional[CssUnit] = None,
    gap: Optional[CssUnit] = None,
    class_: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> RecallContextManager[Tag]:
    """
    Context manager for a grid-like, column-first layout

    This function wraps :func:`~shiny.ui.layout_column_wrap`.

    Wraps a 1d sequence of UI elements into a 2d grid. The number of columns (and rows)
    in the grid dependent on the column `width` as well as the size of the display.

    Parameters
    ----------
    width
        The desired width of each card. It can be one of the following:

        * A (unit-less) number between 0 and 1, specified as `1/num`, where `num`
          represents the number of desired columns.
        * A CSS length unit representing either the minimum (when `fixed_width=False`)
          or fixed width (`fixed_width=True`).
        * `None`, which allows power users to set the `grid-template-columns` CSS
          property manually, either via a `style` attribute or a CSS stylesheet.
        * If missing, a value of `200px` will be used.
    fixed_width
        When `width` is greater than 1 or is a CSS length unit, e.g. `"200px"`,
        `fixed_width` indicates whether that `width` value represents the absolute size
        of each column (`fixed_width=TRUE`) or the minimum size of a column
        (`fixed_width=FALSE`).

        When `fixed_width=FALSE`, new columns are added to a row when `width` space is
        available and columns will never exceed the container or viewport size.

        When `fixed_width=TRUE`, all columns will be exactly `width` wide, which may
        result in columns overflowing the parent container.
    heights_equal
        If `"all"` (the default), every card in every row of the grid will have the same
        height. If `"row"`, then every card in _each_ row of the grid will have the same
        height, but heights may vary between rows.
    fill
        Whether or not to allow the layout to grow/shrink to fit a fillable container
        with an opinionated height (e.g., :func:`~shiny.ui.page_fillable`).
    fillable
        Whether or not each element is wrapped in a fillable container.
    height,max_height,min_height
        A valid CSS unit (e.g., `height="200px"`). Use `min_height` and `max_height` in
        a filling layout to ensure that the layout container does not shrink below a
        `min_height` or grow beyond a `max_height`.
    height_mobile
        Any valid CSS unit to use for the height when on mobile devices (or narrow
        windows).
    gap
        Any valid CSS unit to use for the gap between columns.
    class_
        A CSS class to apply to the containing element.
    **kwargs
        Additional attributes to apply to the containing element.
    """
    return RecallContextManager(
        ui.layout_column_wrap,
        kwargs=dict(
            width=width,
            fixed_width=fixed_width,
            heights_equal=heights_equal,
            fill=fill,
            fillable=fillable,
            height=height,
            min_height=min_height,
            max_height=max_height,
            height_mobile=height_mobile,
            gap=gap,
            class_=class_,
            **kwargs,
        ),
    )


@add_example()
def layout_columns(
    *,
    col_widths: BreakpointsUser[int] = None,
    row_heights: BreakpointsUser[CssUnit] = None,
    fill: bool = True,
    fillable: bool = True,
    gap: Optional[CssUnit] = None,
    class_: Optional[str] = None,
    height: Optional[CssUnit] = None,
    min_height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    **kwargs: TagAttrValue,
) -> RecallContextManager[Tag]:
    """
    Context manager for responsive, column-based grid layouts, based on a 12-column
    grid.

    This function wraps :func:`~shiny.ui.layout_columns`.

    Parameters
    ----------
    col_widths
        The widths of the columns, possibly at different breakpoints. Can be one of the
        following:

        * `None` (the default): Automatically determines a sensible number of columns
          based on the number of children given to the layout.
        * A list or tuple of integers between 1 and 12, where each element represents
          the number of columns for the relevant UI element. Column widths are recycled
          to extend the values in `col_widths` to match the actual number of items in
          the layout, and children are wrapped onto the next row when a row exceeds 12
          column units. For example, `col_widths=(4, 8, 12)` allocates 4 columns to the
          first element, 8 columns to the second element, and 12 columns to the third
          element (which wraps to the next row). Negative values are also allowed, and
          are treated as empty columns. For example, `col_widths=(-2, 8, -2)` would
          allocate 8 columns to an element (with 2 empty columns on either side).
        * A dictionary of column widths at different breakpoints. The keys should be one
          of `"xs"`, `"sm"`, `"md"`, `"lg"`, `"xl"`, or `"xxl"`, and the values are
          either of the above. For example, `col_widths={"sm": (3, 3, 6), "lg": (4)}`.

    row_heights
        The heights of the rows, possibly at different breakpoints. Can be one of the
        following:

        * A numeric vector, where each value represents the [fractional
          unit](https://css-tricks.com/introduction-fr-css-unit/) (`fr`) height of the
          relevant row. If there are more rows than values provided, the pattern will be
          repeated. For example, `row_heights=(1, 2)` allows even rows to take up twice
          as much space as odd rows.
        * A list of numeric or CSS length units, where each value represents the height
          of the relevant row. If more rows are needed than values provided, the pattern
          will repeat. For example, `row_heights=["auto", 1]` allows the height of odd
          rows to be driven my it's contents and even rows to be
          [`1fr`](https://css-tricks.com/introduction-fr-css-unit/).
        * A single string containing CSS length units. In this case, the value is
          supplied directly to `grid-auto-rows`.
        * A dictionary of row heights at different breakpoints, where each key is a
          breakpoint name (one of `"xs"`, `"sm"`, `"md"`, `"lg"`, `"xl"`, or `"xxl"`)
          and where the values may be any of the above options.

    fill
        Whether or not to allow the layout to grow/shrink to fit a fillable container
        with an opinionated height (e.g., :func:`~shiny.ui.page_fillable`).

    fillable
        Whether or not each element is wrapped in a fillable container.

    gap
        Any valid CSS unit to use for the gap between columns.

    class_
        CSS class(es) to apply to the containing element.

    height,max_height,min_height
        A valid CSS unit (e.g., `height="200px"`). Use `min_height` and `max_height` in
        a filling layout to ensure that the layout container does not shrink below a
        `min_height` or grow beyond a `max_height`.

    **kwargs
        Additional attributes to apply to the containing element.

    Returns
    -------
    :
        An :class:`~htmltools.Tag` element.

    See Also
    --------
    * :func:`~shiny.express.ui.layout_column_wrap` for laying out elements into a
      uniform grid.

    Reference
    --------
    * [Bootstrap CSS Grid](https://getbootstrap.com/docs/5.3/layout/grid/)
    * [Bootstrap Breakpoints](https://getbootstrap.com/docs/5.3/layout/breakpoints/)
    """
    return RecallContextManager(
        ui.layout_columns,
        kwargs=dict(
            col_widths=col_widths,
            row_heights=row_heights,
            fill=fill,
            fillable=fillable,
            gap=gap,
            class_=class_,
            height=height,
            min_height=min_height,
            max_height=max_height,
            **kwargs,
        ),
    )


@add_example()
def card(
    *,
    full_screen: bool = False,
    height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    min_height: Optional[CssUnit] = None,
    fill: bool = True,
    class_: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> RecallContextManager[Tag]:
    """
    Context manager for Bootstrap card component

    This function wraps :func:`~shiny.ui.card`.

    A general purpose container for grouping related UI elements together with a border
    and optional padding. To learn more about `card()`s, see [this
    article](https://rstudio.github.io/bslib/articles/cards.html).

    Parameters
    ----------
    full_screen
        If `True`, an icon will appear when hovering over the card body. Clicking the
        icon expands the card to fit viewport size.
    height,max_height,min_height
        Any valid CSS unit (e.g., `height="200px"`). Doesn't apply when a card is made
        `full_screen` (in this case, consider setting a `height` in
        :func:`~shiny.experimental.ui.card_body`).
    fill
        Whether or not to allow the card to grow/shrink to fit a fillable container with
        an opinionated height (e.g., :func:`~shiny.ui.page_fillable`).
    class_
        Additional CSS classes for the returned Tag.
    **kwargs
        HTML attributes on the returned Tag.
    """

    # wrapper
    #     A function (which returns a UI element) to call on unnamed arguments in `*args`
    #     which are not already card item(s) (like :func:`~shiny.ui.card_header`,
    #     :func:`~shiny.experimental.ui.card_body`, etc.). Note that non-card items are
    #     grouped together into one `wrapper` call (e.g. given `card("a", "b",
    #     card_body("c"), "d")`, `wrapper` would be called twice, once with `"a"` and
    #     `"b"` and once with `"d"`).

    return RecallContextManager(
        ui.card,
        kwargs=dict(
            full_screen=full_screen,
            height=height,
            max_height=max_height,
            min_height=min_height,
            fill=fill,
            class_=class_,
            **kwargs,
        ),
    )


@add_example()
def card_body(
    *,
    fillable: bool = True,
    min_height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    max_height_full_screen: Optional[CssUnit] | MISSING_TYPE = MISSING,
    height: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    gap: Optional[CssUnit] = None,
    fill: bool = True,
    class_: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> RecallContextManager[CardItem]:
    # For a general overview of the :func:`~shiny.ui.card` API, see [this article](https://rstudio.github.io/bslib/articles/cards.html).
    """
    Card body container

    A general container for the "main content" of a :func:`~shiny.ui.card`. This
    component is designed to be provided as direct children to :func:`~shiny.ui.card`.

    Parameters
    ----------
    *args
        Contents to the card's body. Or tag attributes that are supplied to the
        resolved :class:`~htmltools.Tag` object.
    fillable
        Whether or not the card item should be a fillable (i.e. flexbox) container.
    min_height,max_height,max_height_full_screen
        Any valid CSS length unit. If `max_height_full_screen` is missing, it is set to
        `max_height`.
    height
        Any valid CSS unit (e.g., `height="200px"`). Doesn't apply when a card is made
        `full_screen` (in this case, consider setting a `height` in
        `card_body()`).
    padding
        Padding to use for the body. This can be a numeric vector
        (which will be interpreted as pixels) or a character vector with valid CSS
        lengths. The length can be between one and four. If one, then that value
        will be used for all four sides. If two, then the first value will be used
        for the top and bottom, while the second value will be used for left and
        right. If three, then the first will be used for top, the second will be
        left and right, and the third will be bottom. If four, then the values will
        be interpreted as top, right, bottom, and left respectively.
    gap
        A CSS length unit defining the `gap` (i.e., spacing) between elements provided
        to `*args`. This argument is only applicable when `fillable = TRUE`.
    fill
        Whether to allow this element to grow/shrink to fit its `card` container.
    class_
        Additional CSS classes for the returned Tag.
    **kwargs
        Additional HTML attributes for the returned Tag.

    Returns
    -------
    :
        A :class:`~shiny.ui.CardItem` object.

    See Also
    --------
    * :func:`~shiny.ui.layout_column_wrap` for laying out multiple cards
        (or multiple columns inside a card).
    * :func:`~shiny.ui.card` for creating a card component.
    * :func:`~shiny.ui.card_header` for creating a header within the card.
    * :func:`~shiny.ui.card_footer` for creating a footer within the card.
    """

    return RecallContextManager(
        ui.card_body,
        kwargs=dict(
            fillable=fillable,
            min_height=min_height,
            max_height=max_height,
            max_height_full_screen=max_height_full_screen,
            height=height,
            padding=padding,
            gap=gap,
            fill=fill,
            class_=class_,
            **kwargs,
        ),
    )


@add_example()
def card_header(
    *args: TagChild | TagAttrs,
    container: TagFunction = ui.tags.div,
    **kwargs: TagAttrValue,
) -> RecallContextManager[CardItem]:
    """
    Context manager for a card header container

    This function wraps :func:`~shiny.ui.card_header`.

    A general container for the "header" of a :func:`~shiny.ui.card`. This component is designed
    to be provided as a direct child to :func:`~shiny.ui.card`.

    The header has a different background color and border than the rest of the card.

    Parameters
    ----------
    *args
        Contents to the header container. Or tag attributes that are supplied to the
        resolved :class:`~htmltools.Tag` object.
    container
        Method for the returned Tag object. Defaults to :func:`~shiny.ui.tags.div`.
    **kwargs
        Additional HTML attributes for the returned Tag.
    """
    return RecallContextManager(
        ui.card_header,
        args=args,
        kwargs=dict(
            container=container,
            **kwargs,
        ),
    )


@add_example()
def card_footer(
    *args: TagChild | TagAttrs,
    **kwargs: TagAttrValue,
) -> RecallContextManager[CardItem]:
    """
    Context manager for a card footer container

    This function wraps :func:`~shiny.ui.card_footer`.


    A general container for the "footer" of a :func:`~shiny.ui.card`. This component is designed
    to be provided as a direct child to :func:`~shiny.ui.card`.

    The footer has a different background color and border than the rest of the card.

    Parameters
    ----------
    *args
        Contents to the footer container. Or tag attributes that are supplied to the
        resolved :class:`~htmltools.Tag` object.
    **kwargs
        Additional HTML attributes for the returned Tag.

    """
    return RecallContextManager(
        ui.card_footer,
        args=args,
        kwargs=kwargs,
    )


@add_example()
def accordion(
    *,
    id: Optional[str] = None,
    open: Optional[bool | str | list[str]] = None,
    multiple: bool = True,
    class_: Optional[str] = None,
    width: Optional[CssUnit] = None,
    height: Optional[CssUnit] = None,
    **kwargs: TagAttrValue,
) -> RecallContextManager[Tag]:
    """
    Context manager for a vertically collapsing accordion.

    This function wraps :func:`~shiny.ui.accordion`.

    Parameters
    ----------
    id
        If provided, you can use `input.id()` in your server logic to determine which of
        the :func:`~shiny.ui.accordion_panel`s are currently active. The
        value will correspond to the :func:`~shiny.ui.accordion_panel`'s
        `value` argument.
    open
        A list of :func:`~shiny.ui.accordion_panel` values to open (i.e.,
        show) by default. The default value of `None` will open the first
        :func:`~shiny.ui.accordion_panel`. Use a value of `True` to open
        all (or `False` to open none) of the items. It's only possible to open more than
        one panel when `multiple=True`.
    multiple
        Whether multiple :func:`~shiny.ui.accordion_panel` can be open at
        once.
    class_
        Additional CSS classes to include on the accordion div.
    width
        Any valid CSS unit; for example, height="100%".
    height
        Any valid CSS unit; for example, height="100%".
    **kwargs
        Attributes to this tag.
    """
    return RecallContextManager(
        ui.accordion,
        kwargs=dict(
            id=id,
            open=open,
            multiple=multiple,
            class_=class_,
            width=width,
            height=height,
            **kwargs,
        ),
    )


@add_example()
def accordion_panel(
    title: TagChild,
    *,
    value: Optional[str] | MISSING_TYPE = MISSING,
    icon: Optional[TagChild] = None,
    **kwargs: TagAttrValue,
) -> RecallContextManager[AccordionPanel]:
    """
    Context manager for single accordion panel.

    This function wraps :func:`~shiny.ui.accordion_panel`.

    Parameters
    ----------
    title
        A title to appear in the :func:`~shiny.ui.accordion_panel`'s header.
    value
        A character string that uniquely identifies this panel. If `MISSING`, the
        `title` will be used.
    icon
        A :class:`~htmltools.Tag` which is positioned just before the `title`.
    **kwargs
        Tag attributes to the `accordion-body` div Tag.
    """
    return RecallContextManager(
        ui.accordion_panel,
        args=(title,),
        kwargs=dict(
            value=value,
            icon=icon,
            **kwargs,
        ),
    )


# ======================================================================================
# Nav components
# ======================================================================================


@add_example()
def navset_tab(
    *,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    header: TagChild = None,
    footer: TagChild = None,
) -> RecallContextManager[NavSet]:
    """
    Context manager for a set of nav items as a tabset.

    This function wraps :func:`~shiny.ui.navset_tab`.

    Parameters
    ----------
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    """
    return RecallContextManager(
        ui.navset_tab,
        kwargs=dict(
            id=id,
            selected=selected,
            header=header,
            footer=footer,
        ),
    )


@add_example()
def navset_pill(
    *,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    header: TagChild = None,
    footer: TagChild = None,
) -> RecallContextManager[NavSet]:
    """
    Context manager for a set of nav items as a pillset.

    This function wraps :func:`~shiny.ui.navset_pill`.

    Parameters
    ----------
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    """
    return RecallContextManager(
        ui.navset_pill,
        kwargs=dict(
            id=id,
            selected=selected,
            header=header,
            footer=footer,
        ),
    )


@add_example()
def navset_underline(
    *,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    header: TagChild = None,
    footer: TagChild = None,
) -> RecallContextManager[NavSet]:
    """
    Context manager for a set of nav items whose active/focused navigation links are
    styled with an underline.

    This function wraps :func:`~shiny.ui.navset_underline`.

    Parameters
    ----------
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    """
    return RecallContextManager(
        ui.navset_underline,
        kwargs=dict(
            id=id,
            selected=selected,
            header=header,
            footer=footer,
        ),
    )


@add_example()
def navset_hidden(
    *,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    header: TagChild = None,
    footer: TagChild = None,
) -> RecallContextManager[NavSet]:
    """
    Context manager for nav contents without the nav items.

    This function wraps :func:`~shiny.ui.navset_hidden`.

    Parameters
    ----------
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    """
    return RecallContextManager(
        ui.navset_hidden,
        kwargs=dict(
            id=id,
            selected=selected,
            header=header,
            footer=footer,
        ),
    )


@add_example()
def navset_card_tab(
    *,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    title: Optional[TagChild] = None,
    sidebar: Optional[ui.Sidebar] = None,
    header: TagChild = None,
    footer: TagChild = None,
) -> RecallContextManager[NavSetCard]:
    """
    Context manager for a set of nav items as a tabset inside a card container.

    This function wraps :func:`~shiny.ui.navset_card_tab`.

    Parameters
    ----------
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    sidebar
        A :class:`shiny.ui.Sidebar` component to display on every :func:`~shiny.ui.nav` page.
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    """
    return RecallContextManager(
        ui.navset_card_tab,
        kwargs=dict(
            id=id,
            selected=selected,
            title=title,
            sidebar=sidebar,
            header=header,
            footer=footer,
        ),
    )


@add_example()
def navset_card_pill(
    *,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    title: Optional[TagChild] = None,
    sidebar: Optional[ui.Sidebar] = None,
    header: TagChild = None,
    footer: TagChild = None,
    placement: Literal["above", "below"] = "above",
) -> RecallContextManager[NavSetCard]:
    """
    Context manager for a set of nav items as a tabset inside a card container.

    This function wraps :func:`~shiny.ui.navset_card_pill`.

    Parameters
    ----------
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    sidebar
        A :class:`shiny.ui.Sidebar` component to display on every :func:`~shiny.ui.nav` page.
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    placement
        Placement of the nav items relative to the content.
    """
    return RecallContextManager(
        ui.navset_card_pill,
        kwargs=dict(
            id=id,
            selected=selected,
            title=title,
            sidebar=sidebar,
            header=header,
            footer=footer,
            placement=placement,
        ),
    )


@add_example()
def navset_card_underline(
    *,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    title: Optional[TagChild] = None,
    sidebar: Optional[ui.Sidebar] = None,
    header: TagChild = None,
    footer: TagChild = None,
    placement: Literal["above", "below"] = "above",
) -> RecallContextManager[NavSetCard]:
    """
    Context manager for a set of nav items as a tabset inside a card container.

    This function wraps :func:`~shiny.ui.navset_card_underline`.

    Parameters
    ----------
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    sidebar
        A :class:`shiny.ui.Sidebar` component to display on every :func:`~shiny.ui.nav` page.
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    placement
        Placement of the nav items relative to the content.
    """
    return RecallContextManager(
        ui.navset_card_underline,
        kwargs=dict(
            id=id,
            selected=selected,
            title=title,
            sidebar=sidebar,
            header=header,
            footer=footer,
            placement=placement,
        ),
    )


@add_example()
def navset_pill_list(
    *,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    header: TagChild = None,
    footer: TagChild = None,
    well: bool = True,
    widths: tuple[int, int] = (4, 8),
) -> RecallContextManager[NavSet]:
    """
    Context manager for a set of nav items as a tabset inside a card container.

    This function wraps :func:`~shiny.ui.navset_pill_list`.

    Parameters
    ----------
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match its
        ``value``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    well
        ``True`` to place a well (gray rounded rectangle) around the navigation list.
    widths
        Column widths of the navigation list and tabset content areas respectively.
    """
    return RecallContextManager(
        ui.navset_pill_list,
        kwargs=dict(
            id=id,
            selected=selected,
            header=header,
            footer=footer,
            well=well,
            widths=widths,
        ),
    )


@add_example()
def navset_bar(
    *,
    title: TagChild,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    sidebar: Optional[ui.Sidebar] = None,
    fillable: bool | list[str] = True,
    gap: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    position: Literal[
        "static-top", "fixed-top", "fixed-bottom", "sticky-top"
    ] = "static-top",
    header: TagChild = None,
    footer: TagChild = None,
    bg: Optional[str] = None,
    # TODO: default to 'auto', like we have in R (parse color via webcolors?)
    inverse: bool = False,
    underline: bool = True,
    collapsible: bool = True,
    fluid: bool = True,
) -> RecallContextManager[NavSetBar]:
    """
    Context manager for a set of nav items as a tabset inside a card container.

    This function wraps :func:`~shiny.ui.navset_bar`.

    Parameters
    ----------
    title
        Title to display in the navbar.
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    sidebar
        A :class:`~shiny.ui.Sidebar` component to display on every
        :func:`~shiny.ui.nav_panel` page.
    fillable
        Whether or not to allow fill items to grow/shrink to fit the browser window. If
        `True`, all `nav()` pages are fillable. A character vector, matching the value
        of `nav()`s to be filled, may also be provided. Note that, if a `sidebar` is
        provided, `fillable` makes the main content portion fillable.
    gap
        A CSS length unit defining the gap (i.e., spacing) between elements provided to
        `*args`.
    padding
        Padding to use for the body. This can be a numeric vector (which will be
        interpreted as pixels) or a character vector with valid CSS lengths. The length
        can be between one and four. If one, then that value will be used for all four
        sides. If two, then the first value will be used for the top and bottom, while
        the second value will be used for left and right. If three, then the first will
        be used for top, the second will be left and right, and the third will be
        bottom. If four, then the values will be interpreted as top, right, bottom, and
        left respectively.
    position
        Determines whether the navbar should be displayed at the top of the page with
        normal scrolling behavior ("static-top"), pinned at the top ("fixed-top"), or
        pinned at the bottom ("fixed-bottom"). Note that using "fixed-top" or
        "fixed-bottom" will cause the navbar to overlay your body content, unless you
        add padding (e.g., ``tags.style("body {padding-top: 70px;}")``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    bg
        Background color of the navbar (a CSS color).
    inverse
        Either ``True`` for a light text color or ``False`` for a dark text color.
    collapsible
        ``True`` to automatically collapse the navigation elements into an expandable
        menu on mobile devices or narrow window widths.
    fluid
        ``True`` to use fluid layout; ``False`` to use fixed layout.
    """
    return RecallContextManager(
        ui.navset_bar,
        kwargs=dict(
            title=title,
            id=id,
            selected=selected,
            sidebar=sidebar,
            fillable=fillable,
            gap=gap,
            padding=padding,
            position=position,
            header=header,
            footer=footer,
            bg=bg,
            inverse=inverse,
            underline=underline,
            collapsible=collapsible,
            fluid=fluid,
        ),
    )


@add_example()
def nav_panel(
    title: TagChild,
    *,
    value: Optional[str] = None,
    icon: TagChild = None,
) -> RecallContextManager[NavPanel]:
    """
    Context manager for nav item pointing to some internal content.

    This function wraps :func:`~shiny.ui.nav`.

    Parameters
    ----------
    title
        A title to display. Can be a character string or UI elements (i.e., tags).
    value
        The value of the item. This is used to determine whether the item is active
        (when an ``id`` is provided to the nav container), programmatically select the
        item (e.g., :func:`~shiny.ui.update_navs`), and/or be provided to the
        ``selected`` argument of the navigation container (e.g.,
        :func:`~shiny.ui.navset_tab`).
    icon
        An icon to appear inline with the button/link.
    """
    return RecallContextManager(
        ui.nav_panel,
        args=(title,),
        kwargs=dict(
            value=value,
            icon=icon,
        ),
    )


@add_example()
def nav_control() -> RecallContextManager[NavPanel]:
    """
    Context manager for a control in the navigation container.

    This function wraps :func:`~shiny.ui.nav_control`.
    """
    return RecallContextManager(ui.nav_control)


@add_example()
def nav_menu(
    title: TagChild,
    *,
    value: Optional[str] = None,
    icon: TagChild = None,
    align: Literal["left", "right"] = "left",
) -> RecallContextManager[NavMenu]:
    """
    Context manager for a menu of nav items.

    This function wraps :func:`~shiny.ui.nav_menu`.

    Parameters
    ----------
    title
        A title to display. Can be a character string or UI elements (i.e., tags).
    value
        The value of the item. This is used to determine whether the item is active
        (when an ``id`` is provided to the nav container), programmatically select the
        item (e.g., :func:`~shiny.ui.update_navs`), and/or be provided to the
        ``selected`` argument of the navigation container (e.g.,
        :func:`~shiny.ui.navset_tab`).
    icon
        An icon to appear inline with the button/link.
    align
        Horizontal alignment of the dropdown menu relative to dropdown toggle.
    """

    return RecallContextManager(
        ui.nav_menu,
        args=(title,),
        kwargs=dict(
            value=value,
            icon=icon,
            align=align,
        ),
    )


# ======================================================================================
# Value boxes
# ======================================================================================
@add_example()
def value_box(
    *,
    showcase: Optional[TagChild] = None,
    showcase_layout: (
        ui._valuebox.SHOWCASE_LAYOUTS_STR | ui.ShowcaseLayout
    ) = "left center",
    full_screen: bool = False,
    theme: Optional[str | ui.ValueBoxTheme] = None,
    height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    min_height: Optional[CssUnit] = None,
    fill: bool = True,
    class_: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> RecallContextManager[Tag]:
    """
    Context manager for a value box

    This function wraps :func:`~shiny.ui.value_box`.

    An opinionated (:func:`~shiny.ui.card`-powered) box, designed for displaying a title
    (the 1st child), value (the 2nd child), and other explanation text (other children,
    if any). Optionally, a `showcase` can provide for context for what the `value`
    represents (for example, it could hold an icon, or even a
    :func:`~shiny.ui.output_plot`).

    Parameters
    ----------
    showcase
        A :class:`~htmltools.Tag` child to showcase (e.g., an icon, a
        :func:`~shiny.ui.output_plot`, etc).
    showcase_layout
        One of `"left center"` (default), `"top right"` or `"bottom"`. Alternatively,
        you can customize the showcase layout options with the
        :func:`~shiny.ui.showcase_left_center`, :func:`~shiny.ui.showcase_top_right()`,
        or :func:`~shiny.ui.showcase_bottom()` functions. Use the options functions when
        you want to control the height or width of the showcase area.
    theme
        The name of a theme (e.g. `"primary"`, `"danger"`, `"purple"`, `"bg-green"`,
        `"text-red"`) for the value box, or a theme constructed with
        :func:`~shiny.ui.value_box_theme`. The theme names provide a convenient way to
        use your app's Bootstrap theme colors as the foreground or background colors of
        the value box. For more control, you can create your own theme with
        :func:`~shiny.ui.value_box_theme` where you can pass foreground and background
        colors directly. Bootstrap supported color themes: `"blue"`, `"purple"`,
        `"pink"`, `"red"`, `"orange"`, `"yellow"`, `"green"`, `"teal"`, and `"cyan"`.
        These colors can be used with `bg-NAME`, `text-NAME`, and
        `bg-gradient-NAME1-NAME2` to change the background, foreground, or use a
        background gradient respectively. If a `theme` string does not start with
        `text-` or `bg-`, it will be auto prefixed with `bg-`.
    full_screen
        If `True`, an icon will appear when hovering over the card body. Clicking the
        icon expands the card to fit viewport size.
    height,max_height,min_height
        Any valid CSS unit (e.g., `height="200px"`). Doesn't apply when a value box is
        made `full_screen`.
    fill
        Whether to allow the value box to grow/shrink to fit a fillable container with
        an opinionated height (e.g., :func:`~shiny.ui.page_fillable`).
    class_
        Utility classes for customizing the appearance of the summary card. Use `bg-*`
        and `text-*` classes (e.g, `"bg-danger"` and `"text-light"`) to customize the
        background/foreground colors.
    **kwargs
        Additional attributes to pass to :func:`~shiny.ui.card`.
    """
    return RecallContextManager(
        ui.value_box,
        kwargs=dict(
            showcase=showcase,
            showcase_layout=showcase_layout,
            full_screen=full_screen,
            theme=theme,
            height=height,
            max_height=max_height,
            min_height=min_height,
            fill=fill,
            class_=class_,
            **kwargs,
        ),
    )


# ======================================================================================
# Panels
# ======================================================================================


@no_example()
def panel_well(**kwargs: TagAttrValue) -> RecallContextManager[Tag]:
    """
    Context manager for a well panel

    This function wraps :func:`~shiny.ui.panel_well`.

    A well panel is a simple container with a border and some padding. It's useful for
    grouping related content together.
    """
    return RecallContextManager(
        ui.panel_well,
        kwargs=dict(
            **kwargs,
        ),
    )


@add_example()
def panel_conditional(
    condition: str,
    **kwargs: TagAttrValue,
) -> RecallContextManager[Tag]:
    """
    Context manager for a conditional panel

    This function wraps :func:`~shiny.ui.panel_conditional`.

    Show UI elements only if a ``JavaScript`` condition is ``true``.

    Parameters
    ----------
    condition
        A JavaScript expression that will be evaluated repeatedly to determine whether
        the panel should be displayed.
    **kwargs
        Attributes to place on the panel tag.

    Note
    ----
    In the JS expression, you can refer to input and output JavaScript objects that
    contain the current values of input and output. For example, if you have an input
    with an id of foo, then you can use input.foo to read its value. (Be sure not to
    modify the input/output objects, as this may cause unpredictable behavior.)

    You are not recommended to use special JavaScript characters such as a period . in
    the input id's, but if you do use them anyway, for example, ``id = "foo.bar"``, you
    will have to use ``input["foo.bar"]`` instead of ``input.foo.bar`` to read the input
    value.

    Tip
    ---
    A more powerful (but slower) way to conditionally show UI content is to use
    :class:`~shiny.render.ui`.
    """
    return RecallContextManager(
        ui.panel_conditional,
        args=(condition,),
        kwargs=dict(**kwargs),
    )


@no_example()
def panel_fixed(
    *,
    top: Optional[str] = None,
    left: Optional[str] = None,
    right: Optional[str] = None,
    bottom: Optional[str] = None,
    width: Optional[str] = None,
    height: Optional[str] = None,
    draggable: bool = False,
    cursor: Literal["auto", "move", "default", "inherit"] = "auto",
    **kwargs: TagAttrValue,
) -> RecallContextManager[TagList]:
    """
    Context manager for a panel of absolutely positioned content.

    This function wraps :func:`~shiny.ui.panel_fixed`.

    This function is equivalent to calling :func:`~shiny.ui.panel_absolute` with
    ``fixed=True`` (i.e., the panel does not scroll with the rest of the page). See
    :func:`~shiny.ui.panel_absolute` for more information.

    Parameters
    ----------
    **kwargs
        Arguments passed along to :func:`~shiny.ui.panel_absolute`.

    See Also
    --------
    * :func:`~shiny.ui.panel_absolute`
    """
    return RecallContextManager(
        ui.panel_fixed,
        kwargs=dict(
            top=top,
            left=left,
            right=right,
            bottom=bottom,
            width=width,
            height=height,
            draggable=draggable,
            cursor=cursor,
            **kwargs,
        ),
    )


@add_example()
def panel_absolute(
    *,
    top: Optional[str] = None,
    left: Optional[str] = None,
    right: Optional[str] = None,
    bottom: Optional[str] = None,
    width: Optional[str] = None,
    height: Optional[str] = None,
    draggable: bool = False,
    fixed: bool = False,
    cursor: Literal["auto", "move", "default", "inherit"] = "auto",
    **kwargs: TagAttrValue,
) -> RecallContextManager[TagList]:
    """
    Context manager for a panel of absolutely positioned content.

    This function wraps :func:`~shiny.ui.panel_absolute`.

    Creates a ``<div>`` tag whose CSS position is set to absolute (or fixed if ``fixed =
    True``). The way absolute positioning works in HTML is that absolute coordinates are
    specified relative to its nearest parent element whose position is not set to static
    (which is the default), and if no such parent is found, then relative to the page
    borders. If you're not sure what that means, just keep in mind that you may get
    strange results if you use this function from inside of certain types of panels.

    Parameters
    ----------
    top
        Distance between the top of the panel, and the top of the page or parent
        container.
    left
        Distance between the left side of the panel, and the left of the page or parent
        container.
    right
        Distance between the right side of the panel, and the right of the page or
        parent container.
    bottom
        Distance between the bottom of the panel, and the bottom of the page or parent
        container.
    width
        Width of the panel.
    height
        Height of the panel.
    draggable
        If ``True``, allows the user to move the panel by clicking and dragging.
    fixed
        Positions the panel relative to the browser window and prevents it from being
        scrolled with the rest of the page.
    cursor
        The type of cursor that should appear when the user mouses over the panel. Use
        ``"move"`` for a north-east-south-west icon, ``"default"`` for the usual cursor
        arrow, or ``"inherit"`` for the usual cursor behavior (including changing to an
        I-beam when the cursor is over text). The default is ``"auto"``, which is
        equivalent to ``"move" if draggable else "inherit"``.
    **kwargs
        Attributes added to the content's container tag.

    Tip
    ----
    The position (``top``, ``left``, ``right``, ``bottom``) and size (``width``,
    ``height``) parameters are all optional, but you should specify exactly two of top,
    bottom, and height and exactly two of left, right, and width for predictable
    results.

    Like most other distance parameters in Shiny, the position and size parameters take
    a number (interpreted as pixels) or a valid CSS size string, such as ``"100px"``
    (100 pixels) or ``"25%"``.

    For arcane HTML reasons, to have the panel fill the page or parent you should
    specify 0 for ``top``, ``left``, ``right``, and ``bottom`` rather than the more
    obvious ``width = "100%"`` and ``height = "100%"``.
    """
    return RecallContextManager(
        ui.panel_absolute,
        kwargs=dict(
            top=top,
            left=left,
            right=right,
            bottom=bottom,
            width=width,
            height=height,
            draggable=draggable,
            fixed=fixed,
            cursor=cursor,
            **kwargs,
        ),
    )


# ======================================================================================
# Tooltips and popovers
# ======================================================================================


@add_example()
def tooltip(
    *,
    id: Optional[str] = None,
    placement: Literal["auto", "top", "right", "bottom", "left"] = "auto",
    options: Optional[dict[str, object]] = None,
    **kwargs: TagAttrValue,
) -> RecallContextManager[Tag]:
    """
    Context manager for a tooltip

    This function wraps :func:`~shiny.ui.tooltip`.

    Display additional information when focusing (or hovering over) a UI element.

    Parameters
    ----------
    id
        A character string. Required to reactively respond to the visibility of the
        tooltip (via the `input[id]` value) and/or update the visibility/contents of the
        tooltip.
    placement
        The placement of the tooltip relative to its trigger.
    options
        A list of additional [Bootstrap
        options](https://getbootstrap.com/docs/5.3/components/tooltips/#options).
    """

    return RecallContextManager(
        ui.tooltip,
        kwargs=dict(
            id=id,
            placement=placement,
            options=options,
            **kwargs,
        ),
    )


@add_example()
def popover(
    *,
    title: Optional[TagChild] = None,
    id: Optional[str] = None,
    placement: Literal["auto", "top", "right", "bottom", "left"] = "auto",
    options: Optional[dict[str, object]] = None,
    **kwargs: TagAttrValue,
) -> RecallContextManager[Tag]:
    """
    Context manager for a popover

    This function wraps :func:`~shiny.ui.popover`.

    Display additional information when clicking on a UI element (typically a
    button).

    Parameters
    ----------
    title
        A title to display in the popover. Can be a character string or UI elements
        (i.e., tags).
    id
        A character string. Required to reactively respond to the visibility of the
        popover (via the `input[id]` value) and/or update the visibility/contents of the
        popover.
    placement
        The placement of the popover relative to its trigger.
    options
        A list of additional [Bootstrap
        options](https://getbootstrap.com/docs/5.3/components/popovers/#options).
    """

    return RecallContextManager(
        ui.popover,
        kwargs=dict(
            title=title,
            id=id,
            placement=placement,
            options=options,
            **kwargs,
        ),
    )



from playwright.sync_api import Page
from playwright.sync_api import expect as playwright_expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_input_action_button_kitchen(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    default = controller.InputActionButton(page, "default")
    default.expect_label("Default button")
    default.expect_disabled(False)
    controller.OutputCode(page, "default_txt").expect_value("Button clicked 0 times")
    default.click()
    controller.OutputCode(page, "default_txt").expect_value("Button clicked 1 times")

    width = controller.InputActionButton(page, "width")
    width.expect_width("200px")

    disabled = controller.InputActionButton(page, "disabled")
    disabled.expect_disabled(True)
    # Disabled button should not have an icon
    playwright_expect(disabled.loc.locator("svg.fa")).to_have_count(0)

    icon = controller.InputActionButton(page, "icon")
    playwright_expect(icon.loc.locator("svg.fa")).to_have_count(1)



"""Low-level reactive components."""

from __future__ import annotations

__all__ = (
    "isolate",
    "invalidate_later",
    "flush",
    "lock",
    "on_flushed",
    "get_current_context",
)

import asyncio
import contextlib
import time
import traceback
import typing
import warnings
from contextvars import ContextVar
from typing import TYPE_CHECKING, Awaitable, Callable, Generator, Optional, TypeVar

from .. import _utils
from .._datastructures import PriorityQueueFIFO
from .._docstring import add_example, no_example
from ..types import MISSING, MISSING_TYPE

if TYPE_CHECKING:
    from ..session import Session

T = TypeVar("T")


class ReactiveWarning(RuntimeWarning):
    pass


# By default warnings are shown once; we want to always show them.
warnings.simplefilter("always", ReactiveWarning)


class Context:
    """A reactive context"""

    def __init__(self) -> None:
        self.id: int = _reactive_environment.next_id()
        self._invalidated: bool = False
        self._invalidate_callbacks: list[Callable[[], None]] = []
        self._flush_callbacks: list[Callable[[], Awaitable[None]]] = []

    def __call__(self) -> typing.ContextManager[None]:
        return _reactive_environment.use_context(self)

    def invalidate(self) -> None:
        """Invalidate this context. It will immediately call the callbacks
        that have been registered with onInvalidate()."""

        if self._invalidated:
            return

        self._invalidated = True

        for cb in self._invalidate_callbacks:
            cb()

        self._invalidate_callbacks.clear()

    def on_invalidate(self, func: Callable[[], None]) -> None:
        """Register a function to be called when this context is invalidated"""
        if self._invalidated:
            func()
        else:
            self._invalidate_callbacks.append(func)

    def add_pending_flush(self, priority: int) -> None:
        """Tell the reactive environment that this context should be flushed the
        next time flushReact() called."""
        _reactive_environment.add_pending_flush(self, priority)

    def on_flush(self, func: Callable[[], Awaitable[None]]) -> None:
        """Register a function to be called when this context is flushed."""
        self._flush_callbacks.append(func)

    async def execute_flush_callbacks(self) -> None:
        """Execute all flush callbacks"""
        for cb in self._flush_callbacks:
            await cb()

        self._flush_callbacks.clear()


class Dependents:
    def __init__(self) -> None:
        self._dependents: dict[int, Context] = {}

    def register(self) -> None:
        ctx: Context = get_current_context()

        if ctx.id in self._dependents:
            # This context is already registered; no need to register it.
            return

        self._dependents[ctx.id] = ctx

        def on_invalidate_cb() -> None:
            if ctx.id in self._dependents:
                del self._dependents[ctx.id]

        ctx.on_invalidate(on_invalidate_cb)

    def invalidate(self) -> None:
        # TODO: Check sort order
        # Invalidate all dependents. This gets all the dependents as list, then iterates
        # over the list. It's done this way instead of iterating over keys because it's
        # possible that a dependent is removed from the dict while iterating over it.
        # https://github.com/posit-dev/py-shiny/issues/26
        ids = sorted(self._dependents.keys())
        for dep_ctx in [self._dependents[id] for id in ids]:
            dep_ctx.invalidate()


class ReactiveEnvironment:
    """The reactive environment"""

    def __init__(self) -> None:
        self._current_context: ContextVar[Optional[Context]] = ContextVar(
            "current_context", default=None
        )
        self._next_id: int = 0
        self._pending_flush_queue: PriorityQueueFIFO[Context] = PriorityQueueFIFO()
        self._lock: Optional[asyncio.Lock] = None
        self._flushed_callbacks = _utils.AsyncCallbacks()

    @property
    def lock(self) -> asyncio.Lock:
        """
        Lock that protects this ReactiveEnvironment. It must be lazily created, because
        at the time the module is loaded, there generally isn't a running asyncio loop
        yet. This causes the asyncio.Lock to be created with a different loop than it
        will be invoked from later; when that happens, acquire() will succeed if there's
        no contention, but throw a "hey you're on the wrong loop" error if there is.
        """
        if self._lock is None:
            # Ensure we have a loop; get_running_loop() throws an error if we don't
            asyncio.get_running_loop()
            self._lock = asyncio.Lock()
        return self._lock

    def next_id(self) -> int:
        """Return the next available id"""
        id = self._next_id
        self._next_id += 1
        return id

    @contextlib.contextmanager
    def use_context(self, ctx: Context) -> typing.Generator[None, None, None]:
        old = self._current_context.set(ctx)
        try:
            yield
        finally:
            self._current_context.reset(old)

    def current_context(self) -> Context:
        """Return the current `Context` object"""
        ctx = self._current_context.get()
        if ctx is None:
            raise RuntimeError("No current reactive context")
        return ctx

    def on_flushed(
        self, func: Callable[[], Awaitable[None]], once: bool = False
    ) -> Callable[[], None]:
        return self._flushed_callbacks.register(func, once=once)

    async def flush(self) -> None:
        """Flush all pending operations"""
        await self._flush_sequential()
        await self._flushed_callbacks.invoke()

    async def _flush_sequential(self) -> None:
        # Sequential flush: instead of storing the tasks in a list and calling gather()
        # on them later, just run each effect in sequence.
        while not self._pending_flush_queue.empty():
            ctx = self._pending_flush_queue.get()
            await ctx.execute_flush_callbacks()

    def add_pending_flush(self, ctx: Context, priority: int) -> None:
        self._pending_flush_queue.put(priority, ctx)

    @contextlib.contextmanager
    def isolate(self) -> Generator[None, None, None]:
        token = self._current_context.set(Context())
        try:
            yield
        finally:
            self._current_context.reset(token)


_reactive_environment = ReactiveEnvironment()


@add_example()
@contextlib.contextmanager
def isolate() -> Generator[None, None, None]:
    """
    Create a non-reactive scope within a reactive scope.

    Ordinarily, the simple act of reading a reactive value causes a relationship to be
    established between the caller and the reactive value, where a change to the
    reactive value will cause the caller to re-execute. (The same applies for the act of
    getting a reactive calculation's value.) `with isolate()` lets you read a reactive
    value or calculation without establishing this relationship.

    ``with isolate()`` can also be useful for calling reactive calculations at the
    console, which can be useful for debugging. To do so, wrap the calls to the reactive
    calculation with ``with isolate()``.

    Returns
    -------
    :
        A context manager that executes the given expression in a scope where reactive
        values can be read, but do not cause the reactive scope of the caller to be
        re-evaluated when they change.

    See Also
    --------
    * :func:`~shiny.reactive.event`
    """
    with _reactive_environment.isolate():
        yield


def get_current_context() -> Context:
    """
    Get the current reactive context.

    Returns
    -------
    :
        A `~shiny.reactive.Context` class.

    Raises
    ------
    RuntimeError
        If called outside of a reactive context.
    """
    return _reactive_environment.current_context()


@no_example()
async def flush() -> None:
    """
    Run any pending invalidations (i.e., flush the reactive environment).

    Warning
    -------
    You shouldn't ever need to call this function inside of a Shiny app. It's only
    useful for testing and running reactive code interactively in the console.
    """
    await _reactive_environment.flush()


@no_example()
def on_flushed(
    func: Callable[[], Awaitable[None]], once: bool = False
) -> Callable[[], None]:
    """
    Register a function to be called when the reactive environment is flushed.

    Parameters
    ----------
    func
        The function to be called when the reactive environment is flushed
    once
        Should the function be run once, and then cleared, or should it
        re-run each time the event occurs.

    Returns
    -------
    :
        A function that can be used to unregister the callback.

    See Also
    --------
    * :func:`~shiny.reactive.flush`
    """

    return _reactive_environment.on_flushed(func, once)


@no_example()
def lock() -> asyncio.Lock:
    """
    A lock that should be held whenever manipulating the reactive graph.

    For example, :func:`~shiny.reactive.lock` makes it safe to set a
    :class:`~reactive.value` and call :func:`~shiny.reactive.flush` from a different
    :class:`~asyncio.Task` than the one that is running the Shiny
    :class:`~shiny.Session`.
    """
    return _reactive_environment.lock


@add_example()
def invalidate_later(
    delay: float, *, session: "MISSING_TYPE | Session | None" = MISSING
) -> None:
    """
    Scheduled Invalidation

    When called from within a reactive context, :func:`~shiny.reactive.invalidate_later`
    schedules the reactive context to be invalidated in the given number of seconds.

    Parameters
    ----------
    delay
        The number of seconds to wait before invalidating.

    Note
    ----
    When called within a reactive function (i.e., :func:`~shiny.reactive.effect`,
    :func:`~shiny.reactive.calc`, :class:`shiny.render.ui`, etc.), that reactive context
    is invalidated (and re-executes) after the interval has passed. The re-execution
    will reset the invalidation flag, so in a typical use case, the object will keep
    re-executing and waiting for the specified interval. It's possible to stop this
    cycle by adding conditional logic that prevents the ``invalidate_later`` from being
    run.
    """

    if isinstance(session, MISSING_TYPE):
        from ..session import get_current_session

        # If no session is provided, autodetect the current session (this
        # could be None if outside of a session).
        session = get_current_session()

    ctx = get_current_context()
    # Pass an absolute time to our subtask, rather than passing the delay directly, in
    # case the subtask doesn't get a chance to start sleeping until a significant amount
    # of time has passed.
    deadline = time.monotonic() + delay

    cancellable = True
    # unsub is used to unsubscribe from session.on_ended when time expires. We don't
    # want a ton of event handler registrations sitting there uselessly, keeping object
    # graphs from being gc'd.
    unsub: Optional[Callable[[], None]] = None

    async def _task(ctx: Context, deadline: float) -> None:
        nonlocal cancellable
        try:
            delay = deadline - time.monotonic()
            try:
                await asyncio.sleep(delay)
            except asyncio.CancelledError:
                # This happens when cancel_task is called due to the session ending, or
                # the context being invalidated due to some other reason. There's no
                # reason for us to keep waiting at that point, as ctx.invalidate() can
                # only be a no-op.
                return

            async with lock():
                # Prevent the ctx.invalidate() from killing our own task. (Another way
                # to accomplish this is to unregister our ctx.on_invalidate handler, but
                # ctx.on_invalidate doesn't currently allow unregistration.)
                cancellable = False

                ctx.invalidate()
                await flush()

        except BaseException:
            traceback.print_exc()
            raise
        finally:
            if unsub:
                unsub()

    task = asyncio.create_task(_task(ctx, deadline))

    def cancel_task():
        if cancellable and not task.cancelled():
            task.cancel()

    ctx.on_invalidate(cancel_task)
    if session:
        unsub = session.on_ended(cancel_task)



from __future__ import annotations

import polars as pl

# pyright: reportMissingTypeStubs = false
from palmerpenguins import load_penguins_raw

from shiny import App, Inputs, render, ui
from shiny.render._data_frame_utils._styles import StyleInfo

# Load the dataset
penguins = load_penguins_raw()
df = penguins

df = df.iloc[0:5, 0:6]


df_styles: list[StyleInfo] = [
    {
        "class": "everywhere",
    },
    {
        "location": "body",
        "rows": [1, 2],
        "cols": "Species",
        "class": "species",
    },
]

app_ui = ui.page_fillable(
    ui.tags.style(
        """
        .everywhere {
            color: darkorange !important;
            font-weight: bold;
        }
        .species {
            background-color: lightblue !important;
        }
        """
    ),
    {"class": "p-3"},
    ui.card(
        ui.card_header("Pandas Styles List:"),
        ui.output_data_frame("pd_list_styles"),
        height="400px",
    ),
    ui.card(
        ui.card_header("Polars Styles List:"),
        ui.output_data_frame("pl_list_styles"),
        height="400px",
    ),
)


def server(input: Inputs):

    @render.data_frame
    def pd_list_styles():
        return render.DataTable(
            df,
            selection_mode=("rows"),
            editable=True,
            # filters=True,
            styles=df_styles,
        )

    @render.data_frame
    def pl_list_styles():

        return render.DataTable(
            pl.from_pandas(df),
            selection_mode=("rows"),
            editable=True,
            # filters=True,
            styles=df_styles,
        )


app = App(app_ui, server, debug=False)



from __future__ import annotations

# Make check to determine if all of shiny.ui is accounted for. Prevent new UI methods from being added without making sure they have module support or could be black listed from the check
# this is a proof of concept of how it can be done.. if the number of items not listed anywhere is > 0 then we fail the test
import datetime
from typing import NamedTuple

from app import input_keys, x_input_keys

# TODO-karan: Verify inputs work within modules
# From https://github.com/posit-dev/py-shiny/pull/731#discussion_r1336102593:
# I like this idea and where it's going...
#
# - [ ] Move `test_0696_resolve_id.py`'s custom expectation method into a new file (`mod_state.py`?) that can be imported into here and the original file.
# - [x] Change class name to `ModState`
# - [ ] Have expectation methods receive a `ModState` object.
# - [ ] Double check if a value is missing when creating a NamedTuple, an error is thrown.
# - [ ] Have a `default_mod_state` object (which currently equals `component` above).
# - [ ] Is it possible to import `mod_state` into the app to retrieve the `input_keys` and `x_input_keys`?


class ModState(NamedTuple):
    sidebar: bool
    accordion: tuple[str, ...]
    popover: bool
    tooltip: bool
    input_action_button: int
    input_action_link: int
    input_file: str | None
    input_checkbox: bool
    input_checkbox_group: tuple[str, ...]
    input_date: datetime.date | str
    input_date_range: tuple[datetime.date, datetime.date]
    input_numeric: int
    input_password: str
    input_radio_buttons: str
    input_select: str
    input_selectize: str
    input_slider: int
    input_switch: bool
    input_text: str
    input_text_area: str
    navset_bar: str
    navset_card_pill: str
    navset_card_tab: str
    navset_hidden: str
    navset_pill: str
    navset_tab: str


blacklist = set(["awesome_component"])

x_input_keys = ("x_" + key for key in x_input_keys)

component = ModState(
    sidebar=True,
    accordion=("a",),
    popover=False,
    tooltip=False,
    input_action_button=0,
    input_action_link=0,
    input_file=None,
    input_checkbox=False,
    input_checkbox_group=("b", "c", "d"),
    input_date=datetime.date(2023, 8, 24),
    input_date_range=(datetime.date(2023, 8, 25), datetime.date(2023, 8, 27)),
    input_numeric=0,
    input_password="password0",
    input_radio_buttons="a",
    input_select="a",
    input_selectize="a",
    input_slider=0,
    input_switch=False,
    input_text="text0",
    input_text_area="text_area0",
    navset_bar="a",
    navset_card_pill="a",
    navset_card_tab="a",
    navset_hidden="a",
    navset_pill="a",
    navset_tab="a",
)

not_in_production = set(component._fields) - set(input_keys)
not_in_experimental = not_in_production - set(x_input_keys)
not_listed = not_in_experimental - blacklist
# TODO-karan: Turn into assertions
print(f" Items not in production are {not_in_production}")
print(f" Items not in experimental are {not_in_experimental}")
print(f"Number of items not listed anywhere are {len(not_listed)}")



from __future__ import annotations

import pkgutil

import palmerpenguins  # pyright: ignore[reportMissingTypeStubs]
import polars as pl
from narwhals.stable.v1.typing import IntoDataFrame

from shiny import Inputs, Outputs, Session
from shiny.express import module, render, ui

pd_penguins = palmerpenguins.load_penguins_raw()[["Sample Number", "Species", "Region"]]
pl_penguins = pl.read_csv(
    pkgutil.get_data(  # pyright: ignore[reportArgumentType]
        "palmerpenguins", "data/penguins-raw.csv"
    )
)[["Sample Number", "Species", "Region"]]


@module
def df_mod(
    input: Inputs,
    output: Outputs,
    session: Session,
    name: str,
    data: IntoDataFrame,
):

    ui.hr()

    ui.h1(f"{name}")

    ui.h2("Column labels w/ and w/o filters")

    ui.h3("With filters")

    @render.data_frame
    def w_filters():
        return render.DataGrid(data, filters=True)

    ui.h3("Without filters")

    @render.data_frame
    def wo_filters():
        return render.DataGrid(data, filters=False)


df_mod("pandas", "pandas", pd_penguins)
df_mod("polars", "polars", pl_penguins)



from pathlib import Path

from shiny import express


def test_is_express_app(tmp_path: Path):
    tmp_file = str(tmp_path / "app.py")

    def write_tmp_file(s: str):
        with open(tmp_file, "w") as f:
            f.write(s)

    write_tmp_file("import shiny.express")
    assert express.is_express_app(tmp_file, None)
    # Check that it works when passing in app_path
    assert express.is_express_app("app.py", str(tmp_path))

    write_tmp_file("# comment\nimport sys\n\nimport shiny.express")
    assert express.is_express_app(tmp_file, None)

    write_tmp_file("import sys\n\nfrom shiny import App, express")
    assert express.is_express_app(tmp_file, None)

    write_tmp_file("import sys\n\nfrom shiny.express import layout, input")
    assert express.is_express_app(tmp_file, None)

    # Shouldn't find in comment
    write_tmp_file("# import shiny.express")
    assert not express.is_express_app(tmp_file, None)

    # Shouldn't find in a string, even if it looks like an import
    write_tmp_file('"""\nimport shiny.express\n"""')
    assert not express.is_express_app(tmp_file, None)

    # Shouldn't recurse into with, if, for, def, etc.
    write_tmp_file("with f:\n  from shiny import express")
    assert not express.is_express_app(tmp_file, None)

    write_tmp_file("if True:\n  import shiny.express")
    assert not express.is_express_app(tmp_file, None)

    write_tmp_file("for i in range(2):\n  import shiny.express")
    assert not express.is_express_app(tmp_file, None)

    write_tmp_file("def f():\n  import shiny.express")
    assert not express.is_express_app(tmp_file, None)

    # Look for magic comment - should override import detection
    write_tmp_file("\n#shiny_mode: core\nfrom shiny.express import ui")
    assert not express.is_express_app(tmp_file, None)

    write_tmp_file("#shiny_mode: express\nfrom shiny import ui")
    assert express.is_express_app(tmp_file, None)



from shiny.express import ui

with ui.navset_card_underline(id="tab"):
    with ui.nav_control():
        ui.a("Shiny", href="https://shiny.posit.co", target="_blank")

    with ui.nav_control():
        ui.a(
            "Learn Shiny",
            href="https://shiny.posit.co/py/docs/overview.html",
            target="_blank",
        )



from __future__ import annotations

from playwright.sync_api import FilePayload, Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_input_file_kitchen(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    file1 = controller.InputFile(page, "file1")

    expect(file1.loc_label).to_have_text("Choose CSV File")
    expect(file1.loc_button).to_have_text("Browse...")

    file1.expect_accept([".csv"])
    file1.expect_button_label("Browse...")
    file1.expect_capture(None)
    # file1.expect_files(None)
    file1.expect_label("Choose CSV File")
    file1.expect_multiple(False)
    file1.expect_placeholder("No file selected")
    file1.expect_width(None)

    file_info: FilePayload = {
        "name": "mtcars.csv",
        "mimeType": "text/csv",
        # from plotnine.data import mtcars
        # mtcars.loc[0:3, mtcars.columns[0:4]].to_csv()
        "buffer": b",name,mpg,cyl,disp\n0,Mazda RX4,21.0,6,160.0\n1,Mazda RX4 Wag,21.0,6,160.0\n2,Datsun 710,22.8,4,108.0\n3,Hornet 4 Drive,21.4,6,258.0\n",
    }
    file_info2 = file_info.copy()
    file_info2["name"] = "mtcars2.csv"

    file1.set(file_info)
    expect(file1.loc_file_display).to_have_value(file_info.get("name"))

    file1.expect_complete()

    output_table = controller.OutputTable(page, "summary")

    output_table.expect_column_labels(["Row Count", "Column Count", "Column Names"])
    output_table.expect_nrow(1)

    file2 = controller.InputFile(page, "file2")
    file2.set([file_info, file_info2])
    expect(file2.loc_file_display).to_have_value("2 files")
    controller.OutputTextVerbatim(page, "file2_info").expect_value(
        """File name: mtcars.csv
File type: text/csv
File size: 129 bytes
---
File name: mtcars2.csv
File type: text/csv
File size: 129 bytes
"""
    )
    # file2.expect_files([file_info, file_info])



from __future__ import annotations

from typing import TypeVar, cast, overload

from htmltools import TagAttrs, TagAttrValue, TagChild, div

TagChildT = TypeVar("TagChildT", bound=TagChild)


# If no children are provided, it will not be able to infer the type of `TagChildT`.
# Using `TagChild`
@overload
def consolidate_attrs(
    *args: TagAttrs,
    **kwargs: TagAttrValue,
) -> tuple[TagAttrs, list[TagChild]]: ...


# Same as original definition
@overload
def consolidate_attrs(
    *args: TagChildT | TagAttrs,
    **kwargs: TagAttrValue,
) -> tuple[TagAttrs, list[TagChildT]]: ...


def consolidate_attrs(
    *args: TagChildT | TagAttrs,
    **kwargs: TagAttrValue,
) -> tuple[TagAttrs, list[TagChildT]]:
    tag = div(*args, **kwargs)

    # `TagAttrs` currently isn't compatible with `htmltools._core.TagAttrDict`
    # https://github.com/posit-dev/py-htmltools/pull/55
    # Convert to a plain dict to avoid getting custom methods from TagAttrDict
    # Cast to `TagAttrs` so that `Tag` functions will accept the dictionary.
    attrs = cast(TagAttrs, dict(tag.attrs))

    # Do not alter children structure (like `TagList` does)
    children = [child for child in args if not isinstance(child, dict)]
    return (attrs, children)


@overload
def trinary(x: None) -> None: ...


@overload
def trinary(x: bool | str) -> str: ...


def trinary(x: bool | str | None) -> None | str:
    if x is None:
        return None
    elif x:
        return "true"
    else:
        return "false"



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_page_navbar_fillable(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    page_navbar_fillable = controller.PageNavbar(page, "page_navbar_fillable")
    page_navbar_fillable._expect_content_text(
        "This page could be used to pick a dataset."
    )
    page_navbar_fillable.expect_fillable(True)
    page_navbar_fillable.expect_gap("300px")



from __future__ import annotations

# Import the custom renderer implementations
from renderers import render_capitalize, render_upper

from shiny import App, Inputs, Outputs, Session, ui

app_ui = ui.page_fluid(
    ui.h1("Capitalization renderer"),
    ui.input_text("caption", "Caption:", "Data summary"),
    "@render_upper: ",
    ui.output_text_verbatim("upper", placeholder=True),
    "@render_upper(): ",
    ui.output_text_verbatim("upper_with_paren", placeholder=True),
    "@render_capitalize: ",
    ui.output_text_verbatim("cap_upper", placeholder=True),
    "@render_capitalize(to='lower'): ",
    ui.output_text_verbatim("cap_lower", placeholder=True),
)


def server(input: Inputs, output: Outputs, session: Session):
    # Hovering over `@render_upper` will display the class documentation
    @render_upper
    def upper():
        return input.caption()

    # Hovering over `@render_upper` will display the class documentation as there is no
    # `__init__()` documentation
    @render_upper()
    def upper_with_paren():
        return input.caption()

    # Hovering over `@render_capitalize` will display the class documentation
    @render_capitalize
    def cap_upper():
        return input.caption()

    # Hovering over `@render_capitalize` will display the `__init__()` documentation
    @render_capitalize(to_case="lower")
    def cap_lower():
        return input.caption()


app = App(app_ui, server)



import re

from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("output_image")


def test_output_image_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    img = controller.OutputImage(page, "image")

    img.expect_inline(False)

    img.expect_img_src(re.compile(r"data:image/png;base64"))
    img.expect_img_height(None)
    img.expect_img_width("100px")
    img.expect_img_alt(None)



# Needed for NotRequired with Python 3.7 - 3.9
# See https://www.python.org/dev/peps/pep-0655/#usage-in-python-3-11
from __future__ import annotations

__all__ = (
    "MISSING",
    "MISSING_TYPE",
    "Jsonifiable",
    "FileInfo",
    "ImgData",
    "SafeException",
    "SilentException",
    "SilentCancelOutputException",
)

from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Dict,
    List,
    Literal,
    NamedTuple,
    Optional,
    Protocol,
    Tuple,
    TypeVar,
    Union,
)

from htmltools import TagChild

from ._docstring import add_example
from ._typing_extensions import NotRequired, TypedDict

if TYPE_CHECKING:
    from matplotlib.figure import Figure


# Sentinel value - indicates a missing value in a function call.
class MISSING_TYPE:
    pass


MISSING: MISSING_TYPE = MISSING_TYPE()


T = TypeVar("T")
ListOrTuple = Union[List[T], Tuple[T, ...]]


# Information about a single file, with a structure like:
#   {'name': 'mtcars.csv', 'size': 1303, 'type': 'text/csv', 'datapath: '/...../mtcars.csv'}
# The incoming data doesn't include 'datapath'; that field is added by the
# FileUploadOperation class.
@add_example(ex_dir="./api-examples/input_file")
class FileInfo(TypedDict):
    """
    Class for information about a file upload.

    See Also
    --------
    * :func:`~shiny.ui.input_file`
    """

    name: str
    """The name of the file being uploaded."""
    size: int
    """The size of the file in bytes."""
    type: str
    """The MIME type of the file."""
    datapath: str
    """The path to the file on the server."""


@add_example(ex_dir="./api-examples/output_image")
class ImgData(TypedDict):
    """
    Return type for :class:`~shiny.render.image`.

    See Also
    --------
    * :class:`~shiny.render.image`
    """

    src: str
    """The ``src`` attribute of the ``<img>`` tag."""
    width: NotRequired[str | float]
    """The ``width`` attribute of the ``<img>`` tag."""
    height: NotRequired[str | float]
    """The ``height`` attribute of the ``<img>`` tag."""
    alt: NotRequired[str]
    """The ``alt`` attribute of the ``<img>`` tag."""
    style: NotRequired[str]
    """The ``style`` attribute of the ``<img>`` tag."""
    coordmap: NotRequired[Any]
    """TODO """


@add_example()
class SafeException(Exception):
    """
    Throw a safe exception.

    When ``shiny.App.SANITIZE_ERRORS`` is ``True`` (which is the case
    in some production environments like Posit Connect), exceptions are sanitized
    to prevent leaking of sensitive information. This class provides a way to
    generate an error that is OK to be displayed to the user.
    """

    pass


@add_example()
class SilentException(Exception):
    """
    Throw a silent exception.

    Normally, when an exception occurs inside a reactive context, it's either:

    - Displayed to the user (as a big red error message)
        - This happens when the exception is raised from an output context (e.g., :class:`shiny.render.ui`)
    - Crashes the application
        - This happens when the exception is raised from an :func:`shiny.reactive.effect`

    This exception is used to silently throw inside a reactive context, meaning that
    execution is paused, and no output is shown to users (or the python console).

    See Also
    --------
    * :class:`~shiny.types.SilentCancelOutputException`
    """

    pass


@add_example()
class SilentCancelOutputException(Exception):
    """
    Throw a silent exception and don't clear output

    Similar to :class:`~shiny.types.SilentException`, but if thrown in an output context,
    existing output isn't cleared.

    See Also
    --------
    * :class:`~shiny.types.SilentException`
    """

    pass


class SilentOperationInProgressException(SilentException):
    # Throw a silent exception to indicate that an operation is in progress

    # Similar to :class:`~SilentException`, but if thrown in an output context, existing
    # output isn't cleared and stays in recalculating mode until the next time it is
    # invalidated.

    pass


class NotifyException(Exception):
    """
    This exception can be raised in a (non-output) reactive effect
    to display a message to the user.

    Parameters
    ----------
    message
        The message to display to the user.
    sanitize
        If ``True``, the message is sanitized to prevent leaking sensitive information.
    close
        If ``True``, the session is closed after the message is displayed.
    """

    sanitize: bool
    close: bool

    def __init__(self, message: str, sanitize: bool = True, close: bool = False):
        super().__init__(message)
        self.sanitize = sanitize
        self.close = close


class ActionButtonValue(int):
    pass


class NavSetArg(Protocol):
    """
    A value suitable for passing to a navigation container (e.g.,
    :func:`~shiny.ui.navset_tab`).
    """

    def resolve(
        self, selected: Optional[str], context: dict[str, Any]
    ) -> tuple[TagChild, TagChild]:
        """
        Resolve information provided by the navigation container.

        Parameters
        ----------
        selected
            The value of the navigation item to be shown on page load.
        context
            Additional context supplied by the navigation container.
        """
        ...

    def get_value(self) -> str | None:
        """
        Get the value of this navigation item (if any).

        This value is only used to determine what navigation item should be shown
        by default when none is specified (i.e., the first navigation item that
        returns a value is used to determine the container's ``selected`` value).
        """
        ...


# =============================================================================
# Types for plots and images
# =============================================================================


# Use this protocol to avoid needing to maintain working stubs for plotnint. If
# good stubs ever become available for plotnine, use those instead.
class PlotnineFigure(Protocol):
    scales: list[Any]
    coordinates: Any
    facet: Any
    layout: Any
    mapping: dict[str, str]
    theme: PlotnineTheme

    def save(
        self,
        filename: BinaryIO,
        format: str,
        units: str,
        dpi: float,
        width: float,
        height: float,
        verbose: bool,
        bbox_inches: object = None,
    ): ...

    def draw(self, show: bool) -> Figure: ...


class PlotnineTheme(NamedTuple):
    themeables: PlotnineThemeables


class PlotnineThemeables(TypedDict):
    figure_size: PlotnineThemeable | None


class PlotnineThemeable(NamedTuple):
    properties: dict[str, Any]


class CoordmapDims(TypedDict):
    width: float
    height: float


class CoordmapPanelLog(TypedDict):
    x: float | None
    y: float | None


class CoordmapPanelDomain(TypedDict):
    left: float
    right: float
    bottom: float
    top: float


class CoordmapPanelRange(TypedDict):
    left: float
    right: float
    bottom: float
    top: float


class CoordmapPanelMapping(TypedDict):
    x: str | None
    y: str | None
    panelvar1: NotRequired[str]
    panelvar2: NotRequired[str]


class CoordmapPanelvarValues(TypedDict):
    panelvar1: NotRequired[float]
    panelvar2: NotRequired[float]


class CoordmapPanel(TypedDict):
    panel: int
    row: NotRequired[int]
    col: NotRequired[int]
    panel_vars: NotRequired[CoordmapPanelvarValues]
    log: CoordmapPanelLog
    domain: CoordmapPanelDomain
    mapping: CoordmapPanelMapping
    range: CoordmapPanelRange


class Coordmap(TypedDict):
    panels: list[CoordmapPanel]
    dims: CoordmapDims


class CoordXY(TypedDict):
    x: float
    y: float


# Data structure sent from client to server when a plot is clicked, double-clicked, or
# hovered.
class CoordInfo(TypedDict):
    x: float
    y: float
    coords_css: CoordXY
    coords_img: CoordXY
    img_css_ratio: CoordXY
    panelvar1: NotRequired[str]
    panelvar2: NotRequired[str]
    mapping: CoordmapPanelMapping
    domain: CoordmapPanelDomain
    range: CoordmapPanelRange
    log: CoordmapPanelLog
    # .nonce: float


class BrushInfo(TypedDict):
    xmin: float
    xmax: float
    ymin: float
    ymax: float
    coords_css: CoordXY
    coords_img: CoordXY
    img_css_ratio: CoordXY
    panelvar1: NotRequired[str]
    panelvar2: NotRequired[str]
    mapping: CoordmapPanelMapping
    domain: CoordmapPanelDomain
    range: CoordmapPanelRange
    log: CoordmapPanelLog
    direction: Literal["x", "y", "xy"]
    # .nonce: float


# https://github.com/python/cpython/blob/df1eec3dae3b1eddff819fd70f58b03b3fbd0eda/Lib/json/encoder.py#L77-L95
# +-------------------+---------------+
# | Python            | JSON          |
# +===================+===============+
# | dict              | object        |
# +-------------------+---------------+
# | list, tuple       | array         |
# +-------------------+---------------+
# | str               | string        |
# +-------------------+---------------+
# | int, float        | number        |
# +-------------------+---------------+
# | True              | true          |
# +-------------------+---------------+
# | False             | false         |
# +-------------------+---------------+
# | None              | null          |
# +-------------------+---------------+
Jsonifiable = Union[
    str,
    int,
    float,
    bool,
    None,
    List["Jsonifiable"],
    Tuple["Jsonifiable"],
    "JsonifiableDict",
]

JsonifiableDict = Dict[str, Jsonifiable]



from __future__ import annotations

import os
import sys
from pathlib import Path

import click
import questionary

from ._main_utils import cli_action, cli_bold, cli_code, path_rel_wd


def add_test_file(
    *,
    app_file: Path | None,
    test_file: Path | None,
):
    if app_file is None:

        def path_exists(x: Path) -> bool | str:
            if not isinstance(x, (str, Path)):
                return False
            if Path(x).is_dir():
                return "Please provide a file path to your Shiny app"
            return Path(x).exists() or f"Shiny app file can not be found: {x}"

        app_file_val = questionary.path(
            "Enter the path to the app file:",
            default=path_rel_wd("app.py"),
            validate=path_exists,
        ).ask()
    else:
        app_file_val = app_file
    # User quit early
    if app_file_val is None:
        sys.exit(1)
    app_file = Path(app_file_val)

    if test_file is None:

        def path_does_not_exist(x: Path) -> bool | str:
            if not isinstance(x, (str, Path)):
                return False
            if Path(x).is_dir():
                return "Please provide a file path for your test file."
            if Path(x).exists():
                return "Test file already exists. Please provide a new file name."
            if not Path(x).name.startswith("test_"):
                return "Test file must start with 'test_'"
            return True

        test_file_val = questionary.path(
            "Enter the path to the test file:",
            default=path_rel_wd(
                os.path.relpath(app_file.parent / "tests" / "test_app.py", ".")
            ),
            validate=path_does_not_exist,
        ).ask()
    else:
        test_file_val = test_file

    # User quit early
    if test_file_val is None:
        sys.exit(1)
    test_file = Path(test_file_val)

    # Make sure app file exists
    if not app_file.exists():
        raise FileExistsError("App file does not exist: ", test_file)
    # Make sure output test file doesn't exist
    if test_file.exists():
        raise FileExistsError("Test file already exists: ", test_file)
    if not test_file.name.startswith("test_"):
        return "Test file must start with 'test_'"

    # if app path directory is the same as the test file directory, use `local_app`
    # otherwise, use `create_app_fixture`
    is_same_dir = app_file.parent == test_file.parent

    test_name = test_file.name.replace(".py", "")
    rel_path = os.path.relpath(app_file, test_file.parent)

    template = (
        f"""\
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def {test_name}(page: Page, local_app: ShinyAppProc):

    page.goto(local_app.url)
    # Add test code here
"""
        if is_same_dir
        else f"""\
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.pytest import create_app_fixture
from shiny.run import ShinyAppProc

app = create_app_fixture("{rel_path}")


def {test_name}(page: Page, app: ShinyAppProc):

    page.goto(app.url)
    # Add test code here
"""
    )
    # Make sure test file directory exists
    test_file.parent.mkdir(parents=True, exist_ok=True)

    # Write template to test file
    test_file.write_text(template)

    # next steps
    click.echo()
    click.echo(cli_action(cli_bold("Next steps:")))
    click.echo(f"- Run {cli_code('pytest')} in your terminal to run all the tests")



import pytest

from shiny._deprecated import ShinyDeprecationWarning
from shiny.ui import div, layout_column_wrap

X = div("42")
Y = div("43")
w = 1 / 2


def test_layout_column_width_as_first_param_is_deprecated():
    layout_column_wrap(X)
    with pytest.warns(ShinyDeprecationWarning, match="`width` parameter must be named"):
        layout_column_wrap(w, X)
    layout_column_wrap(X, w, Y)
    layout_column_wrap(X, Y, w)
    layout_column_wrap(X, Y, width=w)
    layout_column_wrap(X, Y, width=None)



from __future__ import annotations

from playwright.sync_api import Locator, Page
from playwright.sync_api import expect as playwright_expect

from .._types import PatternOrStr, StyleValue, Timeout
from ..expect import expect_not_to_have_class
from ..expect._internal import expect_class_to_have_value as _expect_class_to_have_value
from ..expect._internal import expect_style_to_have_value as _expect_style_to_have_value
from ._base import UiWithContainer, WidthLocM
from ._expect import expect_locator_values_in_list


class AccordionPanel(
    WidthLocM,
    UiWithContainer,
):
    """
    Controller for :func:`shiny.ui.accordion_panel`.
    """

    loc_label: Locator
    """
    Playwright `Locator` for the accordion panel's label.
    """
    loc_icon: Locator
    """
    Playwright `Locator` for the accordion panel's icon.
    """
    loc_body: Locator
    """
    Playwright `Locator` for the accordion panel's body.
    """
    loc_header: Locator
    """
    Playwright `Locator` for the accordion panel's header.
    """
    # loc_body_visible: Locator
    # """
    # Playwright `Locator` for the visible accordion panel body
    # """

    def __init__(self, page: Page, id: str, data_value: str) -> None:
        """
        Initializes a new instance of the `AccordionPanel` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the accordion panel.
        data_value
            The data value of the accordion panel.
        """
        super().__init__(
            page,
            id=id,
            loc=f"> div.accordion-item[data-value='{data_value}']",
            loc_container=f"div#{id}.accordion.shiny-bound-input",
        )

        self.loc_label = self.loc.locator(
            "> .accordion-header > .accordion-button > .accordion-title"
        )

        self.loc_icon = self.loc.locator(
            "> .accordion-header > .accordion-button > .accordion-icon"
        )

        self.loc_body = self.loc.locator("> .accordion-collapse")
        self.loc_header = self.loc.locator("> .accordion-header")
        self._loc_body_visible = self.loc.locator("> .accordion-collapse.show")

    def expect_label(self, value: PatternOrStr, *, timeout: Timeout = None) -> None:
        """
        Expects the accordion panel label to have the specified text.

        Parameters
        ----------
        value
            The expected text pattern or string.
        timeout
            The maximum time to wait for the label to appear. Defaults to `None`.
        """
        playwright_expect(self.loc_label).to_have_text(value, timeout=timeout)

    def expect_body(self, value: PatternOrStr, *, timeout: Timeout = None) -> None:
        """
        Expects the accordion panel body to have the specified text.

        Parameters
        ----------
        value
            The expected text pattern or string.
        timeout
            The maximum time to wait for the body to appear. Defaults to `None`.
        """
        playwright_expect(self.loc_body).to_have_text(value, timeout=timeout)

    def expect_icon(self, exists: bool, *, timeout: Timeout = None) -> None:
        """
        Expects the accordion panel icon to exist or not.

        Parameters
        ----------
        exists
            `True` if the icon is expected to exist, `False` otherwise.
        timeout
            The maximum time to wait for the icon to appear. Defaults to `None`.
        """
        icon_child_loc = self.loc_icon.locator("> *")
        if exists:
            playwright_expect(icon_child_loc).not_to_have_count(0, timeout=timeout)
        else:
            playwright_expect(icon_child_loc).to_have_count(0, timeout=timeout)

    def expect_open(self, value: bool, *, timeout: Timeout = None) -> None:
        """
        Expects the accordion panel to be open or closed.

        Parameters
        ----------
        value
            `True` if the accordion panel is expected to be open, `False` otherwise.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        _expect_class_to_have_value(
            self.loc_body,
            "show",
            has_class=value,
            timeout=timeout,
        )

    # user sends value of Open: true | false
    def set(self, open: bool, *, timeout: Timeout = None) -> None:
        """
        Sets the state of the control to open or closed.

        Parameters
        ----------
        open
            `True` to open the accordion panel, False to close it.
        timeout
            The maximum time to wait for the control to be visible and interactable. Defaults to `None`.
        """
        self.loc.wait_for(state="visible", timeout=timeout)
        self.loc.scroll_into_view_if_needed(timeout=timeout)
        expect_not_to_have_class(self.loc_body, "collapsing", timeout=timeout)
        if self._loc_body_visible.count() != int(open):
            self._toggle(timeout=timeout)

    def _toggle(self, *, timeout: Timeout = None) -> None:
        """
        Toggles the state of the control.

        Parameters
        ----------
        timeout
            The maximum time to wait for the control to be visible and interactable. Defaults to `None`.
        """
        self.loc.wait_for(state="visible", timeout=timeout)
        self.loc.scroll_into_view_if_needed(timeout=timeout)
        self.loc_header.click(timeout=timeout)


class Accordion(
    WidthLocM,
    UiWithContainer,
):
    """Controller for :func:`shiny.ui.accordion`."""

    loc: Locator
    """
    Playwright `Locator` for each accordion items.
    """
    loc_container: Locator
    """
    Playwright `Locator` for the accordion container.
    """
    # loc_open: Locator
    # """
    # `Locator` for the open accordion panel
    # """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `Accordion` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the accordion.
        """
        super().__init__(
            page,
            id=id,
            loc="> div.accordion-item",
            loc_container=f"div#{id}.accordion.shiny-bound-input",
        )
        # self.loc_open = self.loc.locator(
        #   "xpath=.",
        #     # Simple approach as position is not needed
        #     has=page.locator(
        #         "> div.accordion-collapse.show",
        #     ),
        # )

    def expect_height(self, value: StyleValue, *, timeout: Timeout = None) -> None:
        """
        Expects the accordion to have the specified height.

        Parameters
        ----------
        value
            The expected height.
        timeout
            The maximum time to wait for the height to be visible and interactable. Defaults to `None`.
        """
        _expect_style_to_have_value(
            self.loc_container, "height", value, timeout=timeout
        )

    def expect_width(self, value: StyleValue, *, timeout: Timeout = None) -> None:
        """
        Expects the accordion to have the specified width.

        Parameters
        ----------
        value
            The expected width.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_style_to_have_value(self.loc_container, "width", value, timeout=timeout)

    def expect_open(
        self,
        value: list[PatternOrStr],
        *,
        timeout: Timeout = None,
    ) -> None:
        expect_locator_values_in_list(
            page=self.page,
            loc_container=self.loc_container,
            el_type=self.page.locator(
                "> div.accordion-item",
                has=self.page.locator("> div.accordion-collapse.show"),
            ),
            # el_type="> div.accordion-item:has(> div.accordion-collapse.show)",
            arr_name="value",
            arr=value,
            key="data-value",
            timeout=timeout,
            alt_verify=True,
        )

    def expect_panels(
        self,
        value: list[PatternOrStr],
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the accordion to have the specified panels.

        Parameters
        ----------
        value
            The expected panels.
        timeout
            The maximum time to wait for the panels to be visible and interactable. Defaults to `None`.
        """
        expect_locator_values_in_list(
            page=self.page,
            loc_container=self.loc_container,
            el_type="> div.accordion-item",
            arr_name="value",
            arr=value,
            key="data-value",
            timeout=timeout,
            alt_verify=True,
        )

    def set(
        self,
        open: str | list[str],
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Sets the state of the accordion panel.

        Parameters
        ----------
        open
            The open accordion panel(s).
        timeout
            The maximum time to wait for the accordion panel to be visible and interactable. Defaults to `None`.
        """
        if isinstance(open, str):
            open = [open]

        # TODO-future: XOR on the next open state and the current open state
        for i in range(self.loc.count()):
            el_loc = self.loc.nth(i)
            el_loc.element_handle().wait_for_element_state(
                state="visible", timeout=timeout
            )
            el_loc.scroll_into_view_if_needed(timeout=timeout)

            elem_value = el_loc.get_attribute("data-value")
            if elem_value is None:
                raise ValueError(
                    "Accordion panel does not have a `data-value` attribute"
                )
            self.accordion_panel(elem_value).set(elem_value in open, timeout=timeout)

    def expect_class(
        self,
        class_name: str,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the accordion to have the specified class.

        Parameters
        ----------
        class_name
            The class name to expect.
        timeout
            The maximum time to wait for the class to appear. Defaults to `None`.
        """
        _expect_class_to_have_value(
            self.loc_container,
            class_name,
            has_class=True,
            timeout=timeout,
        )

    def expect_multiple(
        self,
        value: bool,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the accordion to be multiple or not.

        Parameters
        ----------
        value
            `True` if the accordion is expected to be multiple, `False` otherwise.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        _expect_class_to_have_value(
            self.loc_container,
            "autoclose",
            has_class=not value,
            timeout=timeout,
        )

    def accordion_panel(
        self,
        data_value: str,
    ) -> AccordionPanel:
        """
        Returns the accordion panel (:class:`~shiny.playwright.controls.AccordionPanel`)
        with the specified data value.

        Parameters
        ----------
        data_value
            The data value of the accordion panel.
        """
        return AccordionPanel(self.page, self.id, data_value)



from pathlib import Path

from shared import filler_text

from shiny import App, render, ui

app_ui = ui.page_sidebar(
    ui.sidebar(
        ui.input_numeric("n", "N", min=0, max=100, value=20),
        title="Parameters",
    ),
    ui.h2("Output"),
    ui.output_text_verbatim("txt"),
    ui.markdown(filler_text),
    title="Theme Example",
    theme=Path(__file__).parent / "css" / "shiny-theme-demo.css",
)


def server(input, output, session):
    @render.text
    def txt():
        return f"n*2 is {input.n() * 2}"


app = App(app_ui, server)



import pandas as pd
import seaborn as sns

from shiny import reactive
from shiny.express import input, render, ui

height = 350
width = "fit-content"
df: reactive.value[pd.DataFrame] = reactive.value(
    sns.load_dataset("anagrams").iloc[:, 1:]
)


def update_data_with_patch(patch):
    df_copy = df().copy()
    fn = str if patch["column_index"] == 0 else int
    df_copy.iat[patch["row_index"], patch["column_index"]] = fn(patch["value"])
    df.set(df_copy)


ui.head_content(
    ui.tags.meta(name="viewport", content="width=device-width, initial-scale=1")
)
ui.input_select(
    "selection_mode",
    "Selection mode",
    {
        "none": "(None)",
        "row": "Single row",
        "rows": "Multiple rows",
    },
    selected="rows",
)
ui.input_switch("filters", "Filters", True)
ui.input_switch("editable", "Editable", True)

with ui.layout_column_wrap(width=1 / 2):
    with ui.card():
        ui.card_header("Data Frame as ", ui.tags.code("render.DataGrid"))

        @render.data_frame
        def grid():
            return render.DataGrid(
                df(),
                width=width,
                height=height,
                filters=input.filters(),
                editable=input.editable(),
                selection_mode=input.selection_mode(),
            )

        @grid.set_patch_fn
        def _(*, patch: render.CellPatch):
            update_data_with_patch(patch)
            return patch["value"]

    with ui.card():
        ui.card_header("Data Frame as ", ui.tags.code("render.DataTable"))

        @render.data_frame
        def table():
            return render.DataTable(
                df(),
                width=width,
                height=height,
                filters=input.filters(),
                editable=input.editable(),
                selection_mode=input.selection_mode(),
            )

        @table.set_patch_fn
        def _(*, patch: render.CellPatch):
            update_data_with_patch(patch)
            return patch["value"]



import pytest
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.render._data_frame_utils._types import ColumnFilterNumber
from shiny.run import ShinyAppProc


@pytest.mark.parametrize("tab_name", ["pandas", "polars"])
def test_dataframe_methods(page: Page, local_app: ShinyAppProc, tab_name: str) -> None:

    page.goto(local_app.url)

    tab = controller.NavsetUnderline(page, "tab")
    tab.set(tab_name)

    data_frame = controller.OutputDataFrame(page, f"{tab_name}-iris_df")
    data_view_rows = controller.OutputCode(page, f"{tab_name}-data_view_rows")
    data_view_selected_true = controller.OutputCode(
        page, f"{tab_name}-data_view_selected_true"
    )
    data_view_selected_false = controller.OutputCode(
        page, f"{tab_name}-data_view_selected_false"
    )
    cell_selection = controller.OutputCode(page, f"{tab_name}-cell_selection")
    reset_df = controller.InputActionButton(page, f"{tab_name}-reset_df")
    update_sort = controller.InputActionButton(page, f"{tab_name}-update_sort")
    update_filter = controller.InputActionButton(page, f"{tab_name}-update_filter")

    def reset_data_frame():
        reset_df.click()
        data_view_rows.expect_value("(0, 1, 2, 3, 4, 5)")
        data_view_selected_true.expect_value("[]")
        data_view_selected_false.expect_value("[0, 1, 50, 51, 100, 101]")
        cell_selection.expect_value("()")

    # assert value of unsorted table
    reset_data_frame()

    # sort column by number descending
    data_frame.set_sort(0)
    data_view_rows.expect_value("(2, 3, 4, 5, 0, 1)")
    data_view_selected_true.expect_value("[]")
    data_view_selected_false.expect_value("[50, 51, 100, 101, 0, 1]")
    cell_selection.expect_value("()")

    # sort column by number ascending
    data_frame.set_sort({"col": 0, "desc": False})
    data_view_rows.expect_value("(1, 0, 5, 4, 3, 2)")
    data_view_selected_true.expect_value("[]")
    data_view_selected_false.expect_value("[1, 0, 101, 100, 51, 50]")
    cell_selection.expect_value("()")

    # sort column by text ascending
    data_frame.set_sort([4])
    data_view_rows.expect_value("(0, 1, 2, 3, 4, 5)")
    data_view_selected_true.expect_value("[]")
    data_view_selected_false.expect_value("[0, 1, 50, 51, 100, 101]")
    cell_selection.expect_value("()")

    # sort column by text descending
    data_frame.set_sort({"col": 4, "desc": True})
    data_view_rows.expect_value("(4, 5, 2, 3, 0, 1) ")
    data_view_selected_true.expect_value("[]")
    data_view_selected_false.expect_value("[100, 101, 50, 51, 0, 1]")
    cell_selection.expect_value("()")

    reset_data_frame()

    # filter using numbers
    filter_num_range: ColumnFilterNumber = {"col": 0, "value": [6, 6.9]}
    data_frame.set_filter(filter_num_range)
    data_view_rows.expect_value("(3, 4)")
    data_view_selected_true.expect_value("[]")
    data_view_selected_false.expect_value("[51, 100]")
    cell_selection.expect_value("()")
    # filter programmatically
    reset_data_frame()
    update_filter.click()
    data_view_rows.expect_value("(3, 4, 5)")
    data_view_selected_true.expect_value("[]")
    data_view_selected_false.expect_value("[51, 100, 101]")
    cell_selection.expect_value("()")

    data_frame.set_sort(3)
    data_view_rows.expect_value("(4, 5, 3)")
    data_view_selected_true.expect_value("[]")
    data_view_selected_false.expect_value("[100, 101, 51]")
    cell_selection.expect_value("()")

    # select single row
    reset_data_frame()
    data_frame.select_rows([1])
    data_view_rows.expect_value("(0, 1, 2, 3, 4, 5)")
    data_view_selected_true.expect_value("[1]")
    data_view_selected_false.expect_value("[0, 1, 50, 51, 100, 101]")
    cell_selection.expect_value("(1,)")

    reset_data_frame()

    # sort columns programmatically
    update_sort.click()
    data_view_rows.expect_value("(0, 4, 3, 2, 1, 5)")
    data_view_selected_true.expect_value("[]")
    data_view_selected_false.expect_value("[0, 100, 51, 50, 1, 101]")
    cell_selection.expect_value("()")

    # select multiple rows
    data_frame.select_rows([3, 1])  # select rows 3 and 1 of (0, 4, 3, 2, 1, 5)
    data_view_rows.expect_value("(0, 4, 3, 2, 1, 5) ")
    data_view_selected_true.expect_value("[100, 50]")
    data_view_selected_false.expect_value("[0, 100, 51, 50, 1, 101]")
    cell_selection.expect_value("(4, 2)")



from __future__ import annotations

import functools
import sys
from types import TracebackType
from typing import Callable, Generic, Mapping, Optional, Type, TypeVar

from htmltools import MetadataNode, Tag, TagList, wrap_displayhook_handler

from .._typing_extensions import ParamSpec

P = ParamSpec("P")
R = TypeVar("R")
U = TypeVar("U")


class RecallContextManager(Generic[R]):
    def __init__(
        self,
        fn: Callable[..., R],
        *,
        args: tuple[object, ...] | None = None,
        kwargs: Mapping[str, object] | None = None,
    ):
        self.fn = fn
        if args is None:
            args = tuple()
        if kwargs is None:
            kwargs = {}
        self.args: list[object] = list(args)
        self.kwargs: dict[str, object] = dict(kwargs)
        # Let htmltools.wrap_displayhook_handler decide what to do with objects before
        # we append them.
        self.wrapped_append = wrap_displayhook_handler(self.args.append)

    def __enter__(self) -> None:
        self._prev_displayhook = sys.displayhook
        sys.displayhook = self.displayhook

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> bool:
        sys.displayhook = self._prev_displayhook
        if exc_type is None:
            res = self.fn(*self.args, **self.kwargs)
            sys.displayhook(res)
        return False

    def displayhook(self, x: object) -> None:
        if isinstance(x, RecallContextManager):
            # This displayhook first checks if x (the child) is a RecallContextManager,
            # in which case it uses `with x` to trigger x.__enter__() and x.__exit__().
            # When x.__exit__() is called, it will invoke x.fn() and then pass the
            # result to this object's (the parent) self.displayhook(), which is this
            # same function, but instead of passing in a RecallContextManager, it will
            # pass in the actual object.
            #
            # In short, this is a way of invoking a re-entrant call to the current
            # function, but instead of passing in a RecallContextManager, it passes in
            # the result from the RecallContextManager.
            with x:
                pass
        else:
            self.wrapped_append(x)

    def tagify(self) -> Tag | TagList | MetadataNode | str:
        res = self.fn(*self.args, **self.kwargs)

        if callable(getattr(res, "tagify", None)):
            return res.tagify()  # pyright: ignore
        if callable(getattr(res, "_repr_html_", None)):
            return res._repr_html_()  # pyright: ignore

        raise RuntimeError(
            "RecallContextManager was used without `with`. When used this way, the "
            "result must have a .tagify() or ._repr_html_() method, but it does not."
        )


def wrap_recall_context_manager(
    fn: Callable[P, R]
) -> Callable[P, RecallContextManager[R]]:
    @functools.wraps(fn)
    def wrapped_fn(*args: P.args, **kwargs: P.kwargs) -> RecallContextManager[R]:
        return RecallContextManager(fn, args=args, kwargs=kwargs)

    return wrapped_fn



import re

from shiny.playwright.expect._expect import (
    _attr_match_str,
    _style_match_str,
    _xpath_match_str,
)


def test_style_match_str() -> None:
    reg = _style_match_str("key", "value")

    def has_match(x: str):
        return reg.search(x) is not None

    def no_match(x: str):
        return reg.search(x) is None

    assert has_match("key: value;")
    assert has_match("key:value")
    assert has_match("key:value;")
    assert has_match(" key : value ; ")

    assert no_match("key:value-bar")
    assert no_match("key:value-bar;")
    assert no_match(" key : value-bar ; ")

    assert no_match("otherkey:value")
    assert no_match("otherkey:value;")
    assert no_match(" otherkey : value ; ")


def test_attr_match_str() -> None:
    assert _attr_match_str("key", "value") == 'key="value"'
    assert _attr_match_str("key", 'value"value2') == r'key="value\"value2"'


def test_xpath_match_str() -> None:
    assert _xpath_match_str("key", "value") == '@key="value"'
    assert _xpath_match_str("key", 'value"value2') == r'@key="value\"value2"'

    assert _xpath_match_str("key", re.compile("value")) == 'matches(@key, "value")'
    assert (
        _xpath_match_str("key", re.compile(r"\bvalue\b"))
        == r'matches(@key, "\bvalue\b")'
    )



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from plotnine.data import mtcars

from shiny import App, Inputs, Outputs, Session, reactive, render, ui

nav = ui.navset_pill_list(
    ui.nav_control(ui.p("Choose a package", class_="lead text-center")),
    ui.nav_panel(
        "Plotnine",
        ui.output_plot("plotnine"),
        ui.div(
            {"class": "d-flex justify-content-center", "style": "gap: 5rem"},
            ui.input_select(
                "x", "X variable", choices=list(mtcars.keys()), selected="wt"
            ),
            ui.input_select(
                "y", "Y variable", choices=list(mtcars.keys()), selected="mpg"
            ),
            ui.input_select(
                "color",
                "Color variable",
                choices=list(mtcars.keys()),
                selected="cyl",
            ),
        ),
    ),
    ui.nav_panel(
        "Seaborn",
        ui.output_plot("seaborn"),
        ui.div(
            {"class": "d-flex justify-content-around"},
            ui.input_slider("var", "Variance", min=0.1, max=10, value=2),
            ui.input_slider("cov", "Co-variance", min=0, max=1, value=0.4),
        ),
    ),
    ui.nav_panel("pandas", ui.output_plot("pandas")),
    ui.nav_panel("Holoviews", ui.output_plot("holoviews", height="600px")),
    ui.nav_panel("xarray", ui.output_plot("xarray")),
    ui.nav_panel("geopandas", ui.output_plot("geopandas")),
    ui.nav_panel("missingno", ui.output_plot("missingno")),
    widths=(2, 10),
    well=False,
)


app_ui = ui.page_fluid(
    ui.panel_title(ui.h2("Py-Shiny static plotting examples", class_="text-center")),
    ui.br(class_="py-3"),
    ui.div(nav, style="max-width: 90%; margin: auto"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.calc
    def fake_data():
        n = 5000
        mean = [0, 0]
        rng = np.random.RandomState(0)
        cov = [(input.var(), input.cov()), (input.cov(), 1 / input.var())]
        return rng.multivariate_normal(mean, cov, n).T

    @render.plot
    def seaborn():
        x, y = fake_data()
        f, ax = plt.subplots(figsize=(6, 6))
        sns.scatterplot(x=x, y=y, s=5, color=".15")
        sns.histplot(x=x, y=y, bins=50, pthresh=0.1, cmap="mako")
        sns.kdeplot(x=x, y=y, levels=5, color="w", linewidths=1)
        return f

    @render.plot
    def plotnine():
        from plotnine import (
            aes,
            facet_wrap,
            geom_point,
            ggplot,
            stat_smooth,
            theme,
            theme_bw,
        )

        color_var = input.color()
        if str(mtcars[color_var].dtype) == "int64":
            color_var = f"factor({color_var})"
        return (
            ggplot(mtcars, aes(input.x(), input.y(), color=color_var))
            + geom_point()
            + stat_smooth(method="lm")
            + facet_wrap("~gear")
            + theme_bw(base_size=16)
            + theme(legend_position="top")
        )

    @render.plot
    def pandas():
        ts = pd.Series(
            np.random.randn(1000), index=pd.date_range("1/1/2000", periods=1000)
        )
        ts = ts.cumsum()
        return ts.plot()

    @render.plot
    def holoviews():
        import holoviews as hv
        from bokeh.sampledata.les_mis import data as les_mis

        links = pd.DataFrame(les_mis["links"])
        return hv.render(hv.Chord(links), backend="matplotlib")

    @render.plot
    def xarray():
        import xarray as xr

        airtemps = xr.tutorial.open_dataset("air_temperature")
        air = airtemps.air - 273.15
        air.attrs = airtemps.air.attrs
        air.attrs["units"] = "deg C"
        return air.isel(lon=10, lat=[19, 21, 22]).plot.line(x="time")

    @render.plot
    def geopandas():
        import geodatasets
        import geopandas

        nybb_path = geodatasets.get_path("nybb")
        boros = geopandas.read_file(nybb_path)
        boros.set_index("BoroCode", inplace=True)
        boros.sort_index(inplace=True)
        return boros.plot()

    @render.plot
    def missingno():
        import matplotlib.pyplot as plt
        import missingno as msno

        collisions = pd.read_csv(
            "https://raw.githubusercontent.com/ResidentMario/missingno-data/master/nyc_collision_factors.csv"
        )
        ret = msno.matrix(
            collisions.sample(250),
            fontsize=8,
            label_rotation=45,
        )
        plt.subplots_adjust(top=0.6)
        return ret


app = App(app_ui, server)



from custom_component import custom_component, render_custom_component

from shiny import App, ui

app_ui = ui.page_fluid(
    ui.input_slider("n", "Choose a value", 1, 20, 5),
    custom_component("myOutput"),
)


def server(input, output, session):
    @render_custom_component
    def myOutput():
        return input.n()


app = App(app_ui, server)



from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("input_radio_buttons")


def test_input_checkbox_group_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    rb = controller.InputRadioButtons(page, "rb")

    expect(rb.loc_label).to_have_text("Choose one:")
    rb.expect_label("Choose one:")

    rb.expect_choices(["html", "text"])
    rb.expect_choice_labels(["Red Text", "Normal text"])
    rb.expect_selected("html")
    rb.expect_inline(False)

    rb.expect_width(None)

    rb.set("text")

    rb.expect_choices(["html", "text"])
    rb.expect_choice_labels(["Red Text", "Normal text"])
    rb.expect_selected("text")



from pathlib import Path

from shiny.express import ui

css_file = Path(__file__).parent / "css" / "styles.css"

"Almost before we knew it, we had left the ground!!!"

ui.include_css(css_file)

# Style individual elements with an attribute dictionary.
ui.p("Bold text", {"style": "font-weight: bold"})



# Experimental

from ._card import card_image
from ._deprecated import card, card_body, card_title

__all__ = (
    # Deprecated
    "card",
    "card_body",
    "card_title",
    # Still experimental
    "card_image",
)
